#include <cstdio>
#include <cctype>
#include <algorithm>
#include <functional>
#include <unordered_set>

#include "assert.h"

#include "binding.hh"
#include "basfun.hh"
#include "typenameof.hh"
#include "compile.hh"
#include "dprintf.hh"

static char DefVarType[26+1] = "!!!!!!!!!!!!!!!!!!!!!!!!!!"; // = DEFSNG A-Z
static bool DynamicMode = false;
static ScopeP RootScope = new Scope;
static ScopeP CurrentScope = RootScope;
static std::string CurrentSubBaseName;

void SetDefVarType(char begin, char end, char type)
{
    begin = std::toupper(begin);
    end   = std::toupper(end);

    if(begin > end) { std::swap(begin, end); }

    /* TODO */
    fprintf(stderr, "DEF-%c characters %c-%c\n", type, begin, end);

    while(begin<=end)
        DefVarType[begin++ - 'A'] = type;
}

void SetMetaDynamic()
{
    fprintf(stderr, "Meta $DYNAMIC\n");
    DynamicMode = true;
}

void SetMetaStatic()
{
    fprintf(stderr, "Meta $STATIC\n");
    DynamicMode = false;
}

bool IsMetaDynamic() { return DynamicMode; }

bool HasTypeSuffix(const std::string& varname)
{
    return !CalcTypeSuffix(varname, false).empty();
}
std::string CalcTypeSuffix(const std::string& name, bool force)
{
    std::size_t suffix_length = 0;
    while(name.size() > suffix_length)
    {
        char c = name[name.size() - 1 - suffix_length];
        // Possible suffices:
        //   ! single
        //   $ string
        //   # double
        //   % integer
        //   %% _byte
        //   & long
        //   ~% _unsigned integer
        //   ~%% _unsigned _byte
        //   ~& _unsigned long
        if(c == '!' || c == '$' || c == '#') { ++suffix_length; break; }
        if(c == '~') { ++suffix_length; break; }
        if(c == '%' || c == '&') { ++suffix_length; continue; }
        break;
    }
    std::string suffix = name.substr(name.size() - suffix_length);
    if(suffix.empty() && force) suffix += DefVarType[ toupper(name[0]) - 'A' ];
    return suffix;
}

std::string StripTypeSuffix(const std::string& varname)
{
    return varname.substr(0, varname.size() - CalcTypeSuffix(varname,false).size());
}

static TypeType TypeSuffixIntoType(const std::string& type)
{
    #define m(s,a,b,c) (a)+(b)*0x100+(c)*0x10000+(s)*0x1000000
    switch(m(type.size(), type.size() > 0 ? type[0] : 0,
                          type.size() > 1 ? type[1] : 0,
                          type.size() > 2 ? type[2] : 0))
    {
        case m(1,'!',0,0): return TypeType::Single;
        case m(1,'#',0,0): return TypeType::Double;
        case m(1,'$',0,0): return TypeType::String;
        case m(1,'%',0,0): return TypeType::Integer;
        case m(1,'&',0,0): return TypeType::Long;
        case m(2,'~','%',0): return TypeType::UnsignedInteger;
        case m(2,'~','&',0): return TypeType::UnsignedLong;
        case m(2,'%','%',0): return TypeType::Byte;
        case m(3,'~','%','%'): return TypeType::UnsignedByte;
    }
    #undef m
    return TypeType::UserType;
}

TypeP CreateTypeTemplate(const std::string& name)
{
    auto t = TypeSuffixIntoType(CalcTypeSuffix(name, true));
    TypeP result;
    if(t != TypeType::UserType) result = new TypeInfo(t);
    return result;
}

template<typename T>
static void ForEachScope(const T& t)
{
    t(*CurrentScope);
    if(&*CurrentScope != &*RootScope) t(*RootScope);
}

const ScopeP& GetCurrentScope()
{
    return CurrentScope;
}

const ScopeP& GetRootScope()
{
    return RootScope;
}

std::string DeclareAsLabelInCurrentScope(const std::string& varname, bool is_definition)
{
    try {
        /* Find from an existing scope */
        ForEachScope( [&varname] (const Scope& scope)
        {
            auto i = scope.labelnames.find(varname);
            if(i != scope.labelnames.end())
            {
                //if(&scope == &*RootScope && __ctx(is_definition)) throw (long double)-1;
                throw i->second;
            }
        });
    }
    catch(const std::string& res)
    {
        return res;
    }

    std::string name = "L_" + BasUCASE(varname);
    CurrentScope->labelnames[varname] = name;
    CurrentScope->labels[name].defined = false;
    return name;
}

bool AnchorLabelHere(const std::string& l)
{
    /* Find from an existing scope */
    try {
        ForEachScope( [l] (Scope& scope)
        {
            auto i = scope.labels.find(l);
            if(i != scope.labels.end())
            {
                if(i->second.defined)
                {
                    dprintf(dp_declare, "Label %s already defined\n", i->first.c_str());
                    throw false;
                }
                i->second.defined = true;
                i->second.dataptr = global_data.size();
                throw true;
            }
        });
    }
    catch(bool res)
    {
        return res;
    }
    LabelDef& def = CurrentScope->labels[l];
#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Defining label %s\n", l.c_str());
#endif
    def.defined = true;
    def.dataptr = global_data.size();
    return true;
}

UtypeP FindUserType(const std::string& name)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

    try
    {
        ForEachScope( [&basename] (const Scope& scope)
        {
            auto i = scope.types.find(basename);
            if(i != scope.types.end())
            {
                throw i->second;
            }
        });
    } catch(UtypeP p) { return p; }

    return nullptr;
}

bool DefineUserType(const std::string& name, UtypeP type)
{
    if(FindUserType(name)) return false;

    const std::string basename = BasUCASE(StripTypeSuffix(name));

    ScopeP scope = GetRootScope();

    scope->types[basename] = type;

    return true;
}

exprp FindConstant(const std::string& name)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

    try
    {
        ForEachScope( [&basename] (const Scope& scope)
        {
            auto i = scope.constants.find(basename);
            if(i != scope.constants.end())
            {
                throw i->second;
            }
        });
    } catch(exprp p) { return p; }

    return nullptr;
}

int DefineConstant(const std::string& name, exprp p)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

    try
    {
        ForEachScope( [&basename] (const Scope& scope)
        {
            auto i = scope.usage.find(basename);
            if(i != scope.usage.end())
            {
                const TypeFlagSet& tf = i->second;
                if(tf.is_scalar || tf.is_const || tf.is_subroutine) throw 0;
            }
        });
    } catch(int b) { return b; }

    ScopeP scope = GetCurrentScope();

    if(HasTypeSuffix(name))
    {
        TypeP target_type = CreateTypeTemplate(name);
        TypeP orig_type   = p->GetType();
        if(*target_type != *orig_type)
        {
            switch(target_type->Type)
            {
                case TypeType::String:
                {
                    /* If the target type is String, and the expression
                     * is not a String, there's nothing we can do to make
                     * it a String.
                     */
                    return -1;
                }
                /* With numeric types, we can delegate the work to these functions. */
                case TypeType::Byte:
                case TypeType::UnsignedByte:
                case TypeType::Integer: p = new expr_builtin_numeric(CINT, p); break;

                case TypeType::Single:  p = new expr_builtin_numeric(CSNG, p); break;

                case TypeType::UnsignedLong:
                case TypeType::Double:  p = new expr_builtin_numeric(CDBL, p); break;

                case TypeType::UnsignedInteger:
                case TypeType::Long:    p = new expr_builtin_numeric(CLNG, p); break;
                /* If the functions comply, they will return a const, which we check
                 * below. If there's a type mismatch, they will not return a const.
                 */
                default: break;
            }
            p = p->optimized();
            if(!p->is_const()) return -1;
        }
    }
    scope->constants[basename] = p;
    scope->usage[basename].is_const = true;

    return 1;
}

VarContext FindVariableFromScope(const std::string& name, bool match_arrays)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

    VarContext result;
    result.is_global = false;

    ScopeP scope = GetCurrentScope();

recheck_context:;
    auto i = scope->usage.find(basename);
    if(i != scope->usage.end())
    {
        const TypeFlagSet& tf = i->second;

        if(match_arrays ? tf.is_array : tf.is_scalar)
            goto found_context;

        if(match_arrays ? tf.is_global_array : tf.is_global_scalar)
        {
            ScopeP root = GetRootScope();
            if(&*scope != &*root)
            {
                scope = root;
                result.is_global = true;
                goto recheck_context;
            }
            goto found_context;
        }
    }
    scope = nullptr;
found_context:
    if(scope)
    {
        std::string typesuffix = CalcTypeSuffix(name, false);
        /* First attempt to find the right scope using explicit type, if there is one */
    recheck_typesuffix:
        const auto& c = scope->Scopes[(unsigned)TypeSuffixIntoType(typesuffix)][match_arrays];

        auto j = c.find(basename);
        if(j != c.end())
        {
            result.info = j->second;
        }
        else if(typesuffix.empty())
        {
            /* If it doesn't work, tack a type suffix and try again */
            typesuffix += DefVarType[ std::toupper(name[0]) - 'A' ];
            goto recheck_typesuffix;
        }
    }
    return result;
}

TypeP ImplicitlyDefineScalar(const std::string& name, bool is_static, bool is_shared, TypeP type)
{
    if(!type) type = CreateTypeTemplate(name);
#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Implicit scalar: %s\n", name.c_str());
#endif
    return ExplicitlyDefine(name, type, is_static, is_shared, true);
}

TypeP ImplicitlyDefineArray(const std::string& name,
                            unsigned num_dimensions,
                            long starting_index,
                            unsigned ending_index,
                            bool is_static,
                            bool is_shared, TypeP type)
{
#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Implicit array: %s\n", name.c_str());
#endif
    if(!type) type = CreateTypeTemplate(name);
    type->arraytype = IsMetaDynamic() ? TypeInfo::DynamicArray : TypeInfo::StaticArray;
    type->dimensions.resize(num_dimensions,
                            std::make_pair((long)(starting_index),
                                           (unsigned)(ending_index-starting_index+1))
                           );
    return ExplicitlyDefine(name, type, is_static, is_shared, true);
}

TypeP ExplicitlyDefine(const std::string& name, TypeP type,
                       bool is_static,
                       bool is_shared,
                       bool was_implicit)
{
    ctxp res = ExplicitlyDefine_C(name,type,is_static,is_shared, was_implicit);
    if(!res)return 0;
    return res->type;
}

ctxp ExplicitlyDefine_C(const std::string& name, TypeP type,
                        bool is_static,
                        bool is_shared,
                        bool was_implicit)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

#ifdef BINDING_DEBUG
    dprintf(dp_declare, "Defining '%s' as '%s'%s%s\n",
        name.c_str(),
        type->Dump().c_str(),
        is_static ? " (static)":"",
        is_shared ? " (shared)":""
    );
#endif

    bool is_array = type->is_array();
    if(FindVariableFromScope(name, is_array)) return 0;

    // If it's a const or a subroutine in any scope, bail out
    try
    {
        ForEachScope( [&basename,is_array] (const Scope& scope)
        {
            auto i = scope.usage.find(basename);
            if(i != scope.usage.end())
            {
                const TypeFlagSet& tf = i->second;
                if(tf.is_subroutine) throw false;
                if(!is_array && tf.is_const) throw false;
            }
        });
    } catch(bool b) { return 0; }

    ScopeP scope = GetCurrentScope();

    ctxp context = new VarContextDef;
    context->type      = type;
    context->is_static = is_static;
    context->is_shared = is_shared;

    if(is_array)
    {
        if(was_implicit)
        {
            if(type->arraytype != TypeInfo::ParamArray)
                type->arraytype = DynamicMode ? TypeInfo::DynamicArray : TypeInfo::StaticArray;
        }
        scope->usage[basename].is_array  = true;

        if(is_shared) scope->usage[basename].is_global_array = true;
    }
    else
    {
        scope->usage[basename].is_scalar = true;

        if(is_shared) scope->usage[basename].is_global_scalar = true;
    }

    (scope->Scopes[(unsigned)type->Type][is_array]) [basename] = context;

    if(!HasTypeSuffix(name))
    {
        (scope->Scopes[(unsigned)TypeType::UserType][is_array]) [basename] = context;
        /* Explicit type: Add to Utypes,
         * so that it is matched when type suffix
         * is omitted.
         */
    }
    return context;
}

SubFuncP DefineSubFunc(bool is_declare,
                       const std::string& name,
                       TypeP returntype,
                       const std::vector<SubRoutine::ParamSpec>& params)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

    ScopeP scope = GetRootScope();
    TypeFlagSet& typeset = scope->usage[basename];

    bool first_def = !typeset.is_subroutine;

    // Not first definition?
    if(!first_def)
    {
        if(is_declare) return 0;        // DECLARE should be the first mention!
        if(typeset.has_body) return 0;  // Anything is useless after the sub has a body
    }

    if(typeset.is_const || typeset.is_scalar) return 0;

    if(!returntype && HasTypeSuffix(name)) return 0; // TODO: Should be caught by parser

    typeset.is_subroutine = true;

    SubFuncP& sub = scope->subroutines[basename];
    if(!sub) sub = new SubRoutine;

    if(first_def)
    {
        sub->returntype = returntype;
        sub->params     = params;
    }
    else
    {
        /* Check that the return type still matches */
        if(!returntype != !sub->returntype
        || (returntype && *sub->returntype  != *returntype))
        {
            fprintf(stderr, "Return type not matching\n");
            return 0;
        }

        /* Check that the parameter types still match */
        for(std::size_t a=0; a<params.size(); ++a)
        {
            if(params[a].passconvention != sub->params[a].passconvention
            && params[a].passconvention != SubRoutine::ParamSpec::ByRef)
            {
                // If BYVAL was used in DECLARE but omitted in SUB, we allow that.
                fprintf(stderr, "Param %u: Passing convention does not match?\n", (unsigned)a);
                return 0;
            }
            if(!params[a].type)
            {
                fprintf(stderr, "Param %u: No type?\n", (unsigned)a);
                return 0;
            }

            if(!sub->params[a].type)
            {
                fprintf(stderr, "Param %u refined as %s\n", (unsigned)a, params[a].type->Dump().c_str());
                sub->params[a].type = params[a].type;
            }
            else if(*params[a].type != *sub->params[a].type)
            {
                fprintf(stderr, "Param %u: Type mismatch? (%s vs %s)\n", (unsigned)a,
                    sub->params[a].type->Dump().c_str(),
                    params[a].type->Dump().c_str());
                return 0;
            }
        }
    }

    for(std::size_t a=0; a<params.size(); ++a)
    {
        const std::string oldname = sub->params[a].name;

        sub->params[a].name = BasUCASE(StripTypeSuffix(params[a].name));

        if(!first_def && oldname != sub->params[a].name)
        {
            fprintf(stderr, "Warning: Param %u name changed from %s to %s\n",
                (unsigned)a,
                oldname.c_str(),
                sub->params[a].name.c_str());
        }
    }

    if(!is_declare)
    {
        /* Check for duplicate parameter names */
        std::unordered_set<std::string> names;
        for(std::size_t a=0; a<params.size(); ++a)
        {
            const std::string& name = sub->params[a].name;
            if(names.find(name) != names.end()) return 0;
            names.insert(name);
        }
    }
    return sub;
}

class SubParamUsageAnalyzer
{
private:
    SubFuncP sub;
    SubParamUsageReport* report;
private:
    void AnalyzeExpr(exprp expr, tristate write_target,
                     SubFuncP callee_func,
                     unsigned callee_param)
    {
        if(!expr) return;

        if(!write_target.is_false())
        {
            /* Maybe or definitely a write target */

            if(expr_var* var = dynamic_cast<expr_var*> (&*expr))
            {
                const std::string& name = var->getname();

                VarContext ctx = FindVariableFromScope(name, var->isarray());

                dprintf(dp_param_tracking, "Write into '%s'%s\n", name.c_str(),
                    write_target.is_maybe() ? " (maybe)" : "");

                if(!ctx)
                {
                    dprintf(dp_param_tracking, "- Just what is '%s' anyway?\n", name.c_str());
                }
                else
                {
                    if(ctx.info->param_index >= 0)
                    {
                        unsigned a = ctx.info->param_index;

                        if(write_target.is_maybe())
                        {
                            dprintf(dp_param_tracking, "- Param %u passed to %p[%u]\n", a,
                                &*callee_func, callee_param);
                            std::pair<SubFuncP, unsigned> data(callee_func, callee_param);
                            report->report[a].PassedAsParam.insert(data);
                        }
                        else
                        {
                            dprintf(dp_param_tracking, "- Param %u written to\n", a);
                            report->report[a].AssignedTo = true;
                        }
                    }
                }
            }
        }
        expr->ForEachSubExpr( [this](exprp e, tristate write_target,
                                     const std::string& name, unsigned param_index)
                                    { /*AnalyzeExpr(e, write_target,
                                                  name.empty() ? nullptr : FindSubFunc(name),
                                                  param_index);*/
                                    } );
    }
    void AnalyzeCode(stmtp stmt)
    {
        if(!stmt) return;

        stmt->ForEachSubStmt( [this](stmtp p) { AnalyzeCode(p); } );
        stmt->ForEachSubExpr( [this](exprp e, tristate write_target,
                                     const std::string& name, unsigned param_index)
                                    { AnalyzeExpr(e, write_target,
                                                  name.empty() ? nullptr : FindSubFunc(name),
                                                  param_index);
                                    } );
    }
public:
    void run_func(SubFuncP subp)
    {
        sub = subp;
        report = &sub->params_usage;

        report->report.clear();
        report->report.resize(sub->params.size());

        AnalyzeCode(subp->code);
    }
};

void EnterSubFuncScope(const std::string& name, bool everything_is_shared)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));
    ScopeP rootscope = GetRootScope();
    assert(&*GetCurrentScope() == &*rootscope);

    fprintf(stderr, "Enter sub '%s' (%s)\n", name.c_str(), basename.c_str());

    TypeFlagSet& typeset = rootscope->usage[basename];
    assert(typeset.is_subroutine);
    assert(!typeset.has_body);
    assert(CurrentSubBaseName.empty());

    CurrentSubBaseName = basename;

    SubFuncP subdata = rootscope->subroutines[basename];

    typeset.has_body = true;

    ScopeP& subscope =
        rootscope->subroutines[basename]->sub_scope = new Scope;

    std::unordered_set<std::string> param_names[2];
    for(const auto& p: subdata->params) param_names[p.is_array].insert(p.name);

    // Import scopes.
    for(const auto& a: rootscope->Scopes) // type => scalars,arrays
        for(unsigned is_array=0; is_array<2; ++is_array)
            for(const auto& t: a[is_array])
                if(t.second->is_shared || everything_is_shared)
                {
                    if(param_names[is_array].find(t.first) != param_names[is_array].end())
                    {
                        fprintf(stderr, "Global (shared) %s variable '%s' shadowed by parameter by same name\n",
                            is_array ? "array" : "scalar",
                            t.first.c_str());
                        continue;
                    }

                    if(is_array)
                    {
                        //fprintf(stderr, "Imports global array '%s'\n", t.first.c_str());
                        subscope->usage[t.first].is_global_array = true;
                    }
                    else
                    {
                        //fprintf(stderr, "Imports global scalar '%s'\n", t.first.c_str());
                        subscope->usage[t.first].is_global_scalar = true;
                    }
                }
    CurrentScope = subscope;

    for(std::size_t a=0; a<subdata->params.size(); ++a)
    {
        const SubRoutine::ParamSpec& p = subdata->params[a];
        // name, type.. ignore passconvention
        ctxp ctx = ExplicitlyDefine_C(p.name, p.type, false, false);
        ctx->param_index = a;
    }
}

void ExitSubFuncScope(const std::string& name, stmtp code)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));
    ScopeP rootscope = GetRootScope();
    ScopeP curscope = GetCurrentScope();
    assert(&*curscope != &*rootscope);
    assert(basename == CurrentSubBaseName);

    fprintf(stderr, "Exiting sub '%s' (%s)\n", name.c_str(), basename.c_str());

    assert(rootscope->subroutines.find(basename) != rootscope->subroutines.end());
    assert(rootscope->subroutines[basename]);

    SubFuncP sub = rootscope->subroutines[basename];
    sub->code = code;

    CurrentScope = rootscope;
    CurrentSubBaseName.clear();
}

void SubScopeImportSharedVariable(const std::string& name, TypeP type, bool is_array)
{
    assert(!CurrentSubBaseName.empty());

    bool found = false;
    bool implicitly_defined = false;

    // If type was not specified (they're not looking for a particularly typed variable),
    // determine the type from the name.
    if(!type) type = CreateTypeTemplate(name);

try_finding_again:
    for(const auto& a: GetRootScope()->Scopes) // type => scalars,arrays
    {
        auto i = a[is_array].find(name);
        if(i != a[is_array].end())
        {
            if(*type == *i->second->type)
            {
                if(is_array)
                {
                    fprintf(stderr, "Imports global array '%s'\n", name.c_str());
                    CurrentScope->usage[name].is_global_array = true;
                }
                else
                {
                    fprintf(stderr, "Imports global scalar '%s'\n", name.c_str());
                    CurrentScope->usage[name].is_global_scalar = true;
                }
                found = true;
            }
        }
    }
    // FIXME TODO
    if(!found)
    {
        if(!implicitly_defined)
        {
            fprintf(stderr, "Warning: Could not find SHARED %s variable by name '%s', type '%s'... Defining!\n",
                is_array ? "array" : "scalar", name.c_str(),
                type->Dump().c_str());

            auto backup = CurrentScope;
            CurrentScope = GetRootScope();

            TypeP deftype = new TypeInfo(*type);
            if(is_array)
                ImplicitlyDefineArray(name,1,0,10, false,false, deftype);
            else
                ImplicitlyDefineScalar(name, false,false, deftype);

            CurrentScope = backup;

            implicitly_defined = true;
            goto try_finding_again;
        }

        fprintf(stderr, "Error: Could not find SHARED %s variable by name '%s', type '%s'!\n",
                is_array ? "array" : "scalar", name.c_str(), type->Dump().c_str());
    }
}

SubFuncP FindSubFunc(const std::string& name)
{
    const std::string basename = BasUCASE(StripTypeSuffix(name));

    ScopeP rootscope = GetRootScope();
    if(rootscope->usage[basename].is_subroutine)
        return rootscope->subroutines[basename];

    return nullptr;
}

const std::string GetCurrentSubBaseName()
{
    return CurrentSubBaseName;
}

void AnalyzeSubParamUsage()
{
    ScopeP old_scope = CurrentScope;

    { SubParamUsageAnalyzer analyzer;
    for(const auto& s: RootScope->subroutines)
    {
        CurrentScope = s.second->sub_scope;
        analyzer.run_func(s.second);
    } }

    /* Optimisticly assume parameters are not modified */
    for(const auto& s: RootScope->subroutines)
    {
        SubRoutine& sub = *s.second;
        for(auto& p: sub.params)
            p.was_modified = false;
    }

    for(;;)
    {
        bool changes = false;
        for(const auto& s: RootScope->subroutines)
        {
            const std::string& subname = s.first;
            SubRoutine& sub = *s.second;
            dprintf(dp_param_tracking, "Checking sub %s\n", subname.c_str());

            for(std::size_t a=0; a<sub.params_usage.report.size(); ++a)
            {
                if(sub.params[a].was_modified) continue;

                SubParamUsageReport::ParamUsage& param_report
                    = sub.params_usage.report[a];
                if(param_report.AssignedTo)
                {
                    dprintf(dp_param_tracking, "%s:%u was modified (assign)\n", subname.c_str(), (unsigned)a);
                    sub.params[a].was_modified = true;
                    changes = true;
                }
                else
                {
                    /* SubFuncP->unsigned param index */
                    for(const auto& p: param_report.PassedAsParam)
                    {
                        if(p.first->params[p.second].was_modified)
                        {
                            dprintf(dp_param_tracking, "%s:%u was modified (proxy)\n",
                                subname.c_str(), (unsigned)a);

                            sub.params[a].was_modified = true;
                            changes = true;
                        }
                    }
                }
            }
        }
        if(!changes) break;
    }
    SetCurrentScope(old_scope);
}

void SetCurrentScope(ScopeP scope)
{
    CurrentScope = scope;
}
