#include <stdint.h>
#include <boost/regex.hpp>
#include <string>
#include <vector>
#include <cstdio>
#include <sstream>
#include <map>

#include "hash.hh"
#include "rangeset.hh"

static bool EXTERNAL_OPERATORS = true;

typedef signed long long IntType;

enum rule_type_t
{
    static_rule,
    dynamic_rule,
    plain_numbers
};

struct IfRangeComparison
{
    unsigned lower; // inclusive
    unsigned upper; // exclusive
    bool is_switchcase;
    bool is_unconditional;
    std::string return_rule;
    unsigned complexity;
};
typedef std::vector<IfRangeComparison> IfRangeSet;

template<typename T>
static const IfRangeSet CreateCaseRule
   (const std::vector<bool>& map,
    std::vector<bool>& dontcareset,
    bool wanted_state,
    rule_type_t rule_type,
    T command,
    unsigned begin=0, unsigned end=~0U)
{
    if(end == ~0U) end = map.size();
    
    unsigned range_first=0, range_last=0;
    bool range_active=false;
    
    const unsigned min_split_distance =
        rule_type == plain_numbers
            ? 12
            : 64-7;
    
    IfRangeSet result;
    
    for(unsigned a=begin; a<end; ++a)
        if(map[a] == wanted_state
        // && !dontcareset[a]
        )
        {
            if(!range_active)
                { newrange: range_first=range_last=a; range_active = true; continue; }
            unsigned distance = a - range_last;
            if(distance >= min_split_distance)
            {
                IfRangeSet tmp =
                    CreateCaseRule(map, dontcareset, wanted_state,
                                   rule_type, command,
                                   range_first, range_last+1);
                result.insert(result.end(), tmp.begin(), tmp.end());
                goto newrange;
            }
            range_last = a;
        }
     
    if(!range_active) IgnoreRule: return result;
    
    begin = range_first;
    end   = range_last + 1;
    
    IfRangeComparison comp;
    comp.lower = begin;
    comp.upper = end;
    comp.is_switchcase = false;
    comp.is_unconditional = false;
    comp.complexity = 0;
    std::stringstream res;
    
    if(rule_type == static_rule)
    {
        bool majorly_matching = false;
        unsigned num_total = end-begin;
        unsigned num_ok    = 0;
        unsigned num_dontcare = 0;
        bool all_dontcare = true;
        for(unsigned a=begin; a<end; ++a)
        {
            if(map[a] == wanted_state)
            {
                ++num_ok;
                if(!dontcareset[a]) all_dontcare = false;
            }
            if(dontcareset[a]) ++num_dontcare;
        }
        if(all_dontcare) goto IgnoreRule;
        
        res << "/* " << num_ok << " matching out of " << num_total;
        if(num_dontcare) res << ", " << num_dontcare << " dontcare";
        res << " */\n";
        
        if(num_ok >= num_total / 2)
        {
            majorly_matching = true;
            wanted_state = !wanted_state;
            num_ok = num_total - num_ok;
        }
        
        rangeset<unsigned> ranges;
        for(unsigned a=begin; a<end; ++a)
        {
            if(dontcareset[a])
            {
                if(wanted_state) goto r1;
                else continue;
            }
            if(map[a] == wanted_state)
            {
              r1:
                ranges.insert(a);
            }
        }
        
        if(!ranges.empty())
        {
            if(ranges.size() <= 3)
            {
                rangeset<unsigned>::const_iterator i = ranges.begin();
                
                if(i->lower > begin || i->upper < end)
                {
                    res << "if(";
                    if(majorly_matching) res << '!';
                    res << "(";
                    
                    bool first = true;
                    for(; i != ranges.end(); ++i)
                    {
                        comp.complexity += 4;
                        
                        if(first) first=false; else res << " || ";
                        res << "(index >= " << i->lower
                         << " && index < " << i->upper
                         << ")";
                    }
                    
                    res << "))\n\t";
                }
            }
            else
            {
                unsigned dim = (end-begin + 7) / 8;
                unsigned granu = 8;
                
                // because of aligning, we maybe better round it up anyway
                if(dim&3) dim=(dim+3)&~3;
                
                std::string type = "uint_least8_t", suff="UINT8_C";
                if(!(dim & 1)) { granu=16; type="uint_least16_t"; suff="UINT16_C"; dim >>= 1;
                if(!(dim & 1)) { granu=32; type="uint_least32_t"; suff="UINT32_C"; dim >>= 1;
#if defined(__WORDSIZE) && (__WORDSIZE == 64)
                if(!(dim & 1)) { granu=64; type="uint_least64_t"; suff="UINT64_C"; dim >>= 1;}
#endif
                }}

                if(dim == 1)
                {
                    comp.complexity += 1;

                    uint_least64_t bits = 0;
                    unsigned bitpos=0;
                    for(unsigned a=begin; a<end; ++a)
                    {
                        bits |= (uint_least64_t)(map[a] == wanted_state) << bitpos;
                        ++bitpos;
                        if(map[a] == wanted_state) dontcareset[a] = true;
                    }
                    
                    res << "unsigned bitpos  = (index-" << begin << ")%" << granu << ";\n";
                    res << "if(";
                    if(majorly_matching) res << '!';
                    res << "(" << suff << '(' << bits << ") & (" << suff << "(1) << bitpos)))";
                }
                else
                {
                    comp.complexity += 2;

                    res << "static const " << type << " bitset[" << dim << "] = {\n";
                    uint_least64_t bits = 0;
                    unsigned bitpos=0;
                    for(unsigned a=begin; a<end; ++a)
                    {
                        bits |= (uint_least64_t)(map[a] == wanted_state) << bitpos;
                        if(++bitpos == granu) { res << suff << '(' << bits << "),"; bitpos=bits=0; } 
                    }
                    if(bitpos) { res << suff << '(' << bits << ')'; }
                    res << "};\n";
                    res << "unsigned bitslot = (index-" << begin << ")/" << granu << ";\n";
                    res << "unsigned bitpos  = (index-" << begin << ")%" << granu << ";\n";
                    res << "if(";
                    if(majorly_matching) res << '!';
                    res << "(bitset[bitslot] & (" << suff << "(1) << bitpos)))";
                }
            }
        }
        else
        {
            comp.is_unconditional = true;
        }

        for(unsigned a=begin; a<end; ++a)
            if(map[a] == (wanted_state ^ majorly_matching))
                dontcareset[a] = true;
        res << "return " << command(0) << ";\n";
    }
    else
    {
        unsigned num_total = end-begin;
        
        if(num_total == 1)
        {
            if(dontcareset[begin]) goto IgnoreRule;
            res << "return " << command(begin) << ";\n";
            comp.is_unconditional = true;
            dontcareset[begin] = true;
        }
        else
        {
            rangeset<unsigned> ranges;
            for(unsigned a=begin; a<end; ++a)
                if(map[a] == wanted_state
                ) ranges.insert(a);

            if(rule_type == plain_numbers)
            {
                IntType minvalue = 0, maxvalue = 0;
                bool first = true;
                for(unsigned a=begin; a<end; ++a)
                {
                    std::stringstream tmp;
                    tmp << command(a);
                    IntType value; tmp >> value;
                    if(first || value < minvalue) minvalue=value;
                    if(first || value > maxvalue) maxvalue=value;
                    first = false;
                }
                IntType range = maxvalue - minvalue + 1;
                std::string type;
                IntType limit;
                
                if(range <= 256) { type="uint_least8_t"; limit=256; }
                else if(range <= 65536) { type="uint_least16_t"; limit=65536; }
                else { type="unsigned long"; limit=0; }
                
                if(limit > 0 && minvalue+range <= limit)
                {
                    minvalue=0;
                }
                
                comp.is_unconditional = true;
                
                ++comp.complexity;

                res << "static const " << type << " datatable[" << num_total << "] = {";
                for(unsigned a=begin; a<end; ++a)
                {
                    std::stringstream tmp;
                    tmp << command(a);
                    IntType value; tmp >> value;

                    res << (value - minvalue) << ',';
                    
                    dontcareset[a] = true;
                }
                res << "};\n";
                res << "return datatable[index - " << begin << "] + " << minvalue << ";\n";
            }
            else
            {
                comp.is_switchcase = true;
                
                for(unsigned a=begin; a<end; ++a)
                    if(map[a] == wanted_state)
                    {
                        ++comp.complexity;
                        res << " case " << a << ": return " << command(a) << ";\n";
                        dontcareset[a] = true;
                    }
            }
        }
    }
    comp.return_rule = res.str();
    
    result.push_back(comp);
    return result;
}

struct Table
{
    std::vector<IntType> values;
    std::vector<bool>   handled;
    
    std::string tab_name;
    std::string elem_type;
    
    const IfRangeSet CreateFallback(std::vector<bool>& dontcareset) const
    {
        return CreateCaseRule(handled, dontcareset,
                              false, plain_numbers,
                              [this] (unsigned a) -> unsigned { return values[a]; }
                             );
    }
};
struct Method
{
    std::vector<bool> coverage;
    std::string formula;
    rule_type_t rule_type;
};

static void TranslateTable(Table& tab)
{
    const std::vector<IntType>& values = tab.values;
    const unsigned size                = tab.values.size();
    tab.handled.resize(size, false);
    std::vector<Method> methods;
    
    /* PASS 2: Offset expressions */
    
    //size_t DataAmount = 0;
    
    struct most_freq_test_type
    {
        IntType dummy, dummy2;
        enum { test1, test2, testmul, testand, testor } test_type;

        most_freq_test_type() { }
        
        typedef struct {} Test1Tag;
        most_freq_test_type(IntType v,Test1Tag) : dummy(v),test_type(test1) { }
        typedef struct {} Test2Tag;
        most_freq_test_type(IntType v,Test2Tag) : dummy(v),test_type(test2) { }
        typedef struct {} TestMulTag;
        most_freq_test_type(IntType muls,IntType diffs,TestMulTag)
            : dummy(muls),dummy2(diffs),test_type(testmul) { }
        typedef struct {} TestAndTag;
        most_freq_test_type(IntType ands,IntType diffs,TestAndTag)
            : dummy(ands),dummy2(diffs),test_type(testand) { }
        typedef struct {} TestOrTag;
        most_freq_test_type(IntType ors,IntType diffs,TestOrTag)
            : dummy(ors),dummy2(diffs),test_type(testor) { }

        bool operator() (unsigned a,IntType val) const
        {
            switch(test_type)
            {
                case test1: return val == (a + dummy);
                case test2: return val == dummy;
                case testmul: return val == (a*dummy + dummy2);
                case testand: return val == (( a*dummy2 ) & dummy);
                case testor: return val == (( a*dummy2 ) | dummy);
            }
            return false;
        }
    };
    
    hash_set<std::string> done_sets;
    for(;;)
    {
        std::string most_freq_val;
        most_freq_test_type most_freq_test;

        unsigned most_freq=0;
        
        {std::map<IntType, unsigned> offs_freqs;
        for(unsigned a=0; a<size; ++a)
        {
            if(tab.handled[a]) continue;

            IntType value = values[a]-a;
            //if(done_sets.find(tmp.str()) != done_sets.end()) continue;

            unsigned freq = ++offs_freqs[value];
            if(freq > most_freq)
            {
                std::stringstream tmp; tmp << "index + " << value;
                
                most_freq_val  = tmp.str();
                most_freq_test = most_freq_test_type(value, most_freq_test_type::Test1Tag());
                most_freq = freq;
            }
        }}

        {std::map<IntType, unsigned> val_freqs;
        for(unsigned a=0; a<size; ++a)
        {
            if(tab.handled[a]) continue;
            IntType value = values[a];
            //if(done_sets.find(tmp.str()) != done_sets.end()) continue;

            unsigned freq = ++val_freqs[value];
            if(freq > most_freq)
            {
                std::stringstream tmp; tmp << value;
                IntType dummy=0;
                most_freq_val  = tmp.str();
                most_freq_test = most_freq_test_type(value, most_freq_test_type::Test2Tag());
                most_freq = freq;
            }
        }}

        ///*std::set<std::string>&,
        auto DoTesting =
            [&,size]
            (const std::string& val,
             const most_freq_test_type& test)
        {
            //bool needs_check = (done_sets).find(val) != (done_sets).end();
            
            unsigned freq = 0;
            for(unsigned a=0; a < size; ++a)
            {
                if(tab.handled[a]) continue;
                if(test(a, tab.values[a])) ++freq;
            }
            if(freq > most_freq)
            {
                most_freq      = freq;
                most_freq_val  = val;
                most_freq_test = test;
            }
        };
        
        for(IntType diffs=-100; diffs<=100; ++diffs)
        for(IntType  muls=2; muls<=4; ++muls)
        {
            std::stringstream s;
            s << "index * " << muls << " + " << diffs;
            DoTesting(s.str(),
                      most_freq_test_type(muls,diffs, most_freq_test_type::TestMulTag()));
        }
#if 0
        for(IntType diffs=0; diffs<=60; ++diffs)
        for(IntType ands=1; ands<=256; ands<4 ? ++ands : ands<<=1)
        {
            std::stringstream s;
            s << "(index + " << diffs << ") & " << ands;
            DoTesting(s.str(),
                      most_freq_test_type(ands,diffs, most_freq_test_type::TestAndTag()));
        }
        for(IntType diffs=0; diffs<=60; ++diffs)
        for(IntType ors=1; ors<=256; ors<4 ? ++ors : ors<<=1)
        {
            std::stringstream s;
            s << "(index + " << diffs << ") | " << ors;
            DoTesting(s.str(),
                      most_freq_test_type(ors,diffs, most_freq_test_type::TestOrTag()));
        }
#endif
        
        if(most_freq < 30) break;
        
        done_sets.insert(most_freq_val);
        
        // 10: 463800
        // 20: 463112
        // 30: 463112
        // 40: 463112
        // 50: 463352
        
        Method met;
        met.coverage.resize(size);
        for(unsigned a=0; a<size; ++a)
            if(most_freq_test(a, values[a]))
            {
                met.coverage[a] = true;
                tab.handled[a] = true;
            }
        
        met.formula   = most_freq_val;
        met.rule_type = static_rule;
        methods.push_back(met);
    }
    
    IfRangeSet ranges;
    std::vector<bool> covered(size, false);
    if(true)
    {
        IfRangeSet tmp = tab.CreateFallback(covered);
        ranges.insert(ranges.end(), tmp.begin(), tmp.end());
    }
    
    /* Create them in order of complexity. Lower-compexity ones first. */
    std::vector<bool> done(methods.size(), false);
    for(unsigned a=0; a<methods.size(); ++a)
    {
        IfRangeSet best;
        unsigned best_complexity=0, best_idx=0;
        std::vector<bool> best_covered;
        bool first=true;
        for(unsigned b=0; b<methods.size(); ++b)
            if(!done[b])
            {
                std::vector<bool> covered_backup = covered;
                IfRangeSet tmp = 
                    CreateCaseRule(methods[b].coverage, covered_backup,
                                   true, methods[a].rule_type,
                                   [&methods,b] (unsigned) -> const std::string&
                                     { return methods[b].formula; }
                                  );
                unsigned com = 0;
                for(unsigned c=0; c<tmp.size(); ++c)
                    com += tmp[c].complexity;
                if(first || com < best_complexity)
                {
                    best = tmp;
                    best_complexity = com;
                    best_idx = b;
                    best_covered = covered_backup;
                    first=false;
                }
            }
        
        done[best_idx] = true;
        ranges.insert(ranges.end(), best.begin(), best.end());
        covered = best_covered;
    }
    
    std::string code;
    /*
    std::sort(ranges.begin(), ranges.end(),
              [] (const IfRangeComparison&a, const IfRangeComparison& b)
              { return a.lower < b.lower; } );
    */
    
    for(unsigned a=0; a<ranges.size(); ++a)
    {
        const IfRangeComparison& r = ranges[a];
        std::stringstream tmp;
        //tmp << "/* rule " << a << " */";
        
        if(r.is_switchcase)
        {
            tmp << "switch(index) {\n";
            tmp << r.return_rule;
            tmp << "}\n";
        }
        else if(r.upper == r.lower+1)
        {
            tmp << "if(index == " << r.lower << ") {\n" << r.return_rule << "}\n";
        }
        else if(r.lower == 0 && r.upper == size)
        {
            //tmp << "/* all covering rule */";
            tmp << r.return_rule;
            if(r.is_unconditional) { code += tmp.str(); goto all_done; }
        }
        else if(r.lower == 0)
        {
            tmp << "if(index < " << r.upper << ") {\n" << r.return_rule << "}\n";
        }
        else if(r.upper == size)
        {
            tmp << "if(index >= " << r.lower << ") {\n" << r.return_rule << "}\n";
        }
        else
        {
            tmp << "if(";
            if(r.lower > 0) tmp << "index >= " << r.lower << " && ";
            tmp << "index < " << r.upper << ") {\n" << r.return_rule << "}\n";
        }
        code += tmp.str();
    }
    code += "return 0;"; all_done:

    std::printf("static struct %s_generator {\n", tab.tab_name.c_str());

    if(EXTERNAL_OPERATORS)
    {
        std::printf("  %s operator[] (unsigned index) const;\n", tab.elem_type.c_str());
    }
    else
    {
        std::printf("  %s operator[] (unsigned index) const {\n", tab.elem_type.c_str());
        std::printf("  %s\n", code.c_str());
        std::printf("\n}\n");
    }
    std::printf("} %s;\n", tab.tab_name.c_str());

    if(EXTERNAL_OPERATORS)
    {
        std::printf("%s %s_generator::operator[] (unsigned index) const {\n",
            tab.elem_type.c_str(),
            tab.tab_name.c_str());
        std::printf("  %s\n", code.c_str());
        std::printf("\n}\n");
    }
    //std::printf("/* Amount of data: %u bytes */\n", DataAmount);
}

static void TranslateTable(const std::string& type,
                           unsigned dim,
                           const std::string& name,
                           const std::string& contents)
{
    Table tab;
    tab.tab_name = name;
    tab.elem_type = type;
    tab.values.reserve(dim);
    for(const char*s = contents.c_str(); *s; )
    {
        while(*s && (*s == ' ' || *s == '\t' || *s == '\r' || *s == '\n')) ++s;
        
        int base = 10;
        if(*s == '0' && s[1] == 'x') { base=16; s+=2; }
        else if(*s == '0') { base=8; ++s; }
        tab.values.push_back(strtol(s, const_cast<char**>(&s), base));

        while(*s && (*s == ' ' || *s == '\t' || *s == '\r' || *s == '\n')) ++s;
        if(*s == ',') { ++s; continue; }
    }
    tab.values.resize(dim, 0);
    TranslateTable(tab);
}
                           
static void Find(const std::string& pattern)
{
    unsigned type_idx = 0;
    unsigned name_idx = 0;
    unsigned value_idx = 0;
    unsigned dim_idx = 0;
    
    unsigned idx_count=0;
    std::string regex_str;
    for(unsigned a=0; a<pattern.size(); ++a)
    {
        char ch = pattern[a];
        if(ch == '$')
            switch(pattern[++a])
            {
                case 'T':
                    type_idx = idx_count++;
                    regex_str += "([A-Za-z_0-9 ]*)";
                    break;
                case 'N':
                    name_idx = idx_count++;
                    regex_str += "([A-Za-z_][A-Za-z_0-9]*)";
                    break;
                case 'D':
                    dim_idx = idx_count++;
                    regex_str += "([0-9]+)";
                    break;
                case 'A':
                    value_idx = idx_count++;
                    regex_str += "([0-9,\r\n ]+)";
                    break;
            }
        else if(ch == '?')
            regex_str += "[\t\r\n ]*";
        else if(ch == ' ')
            regex_str += "[\t\r\n ]+";
        else
        {
            if(ch == '\\' || ch == '{' || ch == '}'
            || ch == '*' || ch == '^' || ch == '$'
            || ch == '[' || ch == ']'
            || ch == '(' || ch == ')' || ch == '+') regex_str += '\\';
            regex_str += ch;
        }
    }
    
    //printf("regex = '%s'\n", regex_str.c_str());

    boost::regex regex(regex_str);
    
    std::string input;
    while(!std::feof(stdin) && !std::ferror(stdin))
    {
        char Buf[8192];
        int n = std::fread(Buf,1,sizeof(Buf),stdin);
        if(n > 0) input.append(Buf, Buf+n);
    }
    boost::smatch res;
    
    std::string::const_iterator begin = input.begin();
    std::string::const_iterator  end  = input.end();
    
    for(;;)
    {
        if(!boost::regex_search(begin,end, res, regex)) break;
        
        std::fwrite(&*begin, 1, res[0].first-begin,  stdout);
        
        begin = res[0].second;
        
        std::string type = res[type_idx+1].str();
        std::string name = res[name_idx+1].str();
        std::string value= res[value_idx+1].str();
        std::string dim  = res[dim_idx+1].str();
        
        TranslateTable(type, std::atoi(dim.c_str()), name, value);
    }

    std::fwrite(&*begin, 1, end-begin,  stdout);
}

int main()
{
    std::puts("#define __STDC_CONSTANT_MACROS");
    std::puts("#include <stdint.h>");
    Find("static $T $N?[?$D?]?=?{?$A?}?;");
}
