#define PARSER_CLIKE_MAX_CASES 4

#include "blocksetting"

#include "expr.hh"
//#include "dimdef.hh"
//#include "casevalue.hh"
#include "binding.hh"
#include "stmt.hh"

#include "lexer.hh"
#include "tokens.hh"
#include "basfun.hh"
#include "compile.hh"

#include "dprintf.hh"

//#define LEXER_DEBUG
//#define PARSER_DEBUG

#include "parser.hh"
#include "parser_lexiterator.hh"

typedef Lex<QBtoken>::iterator iterator_t;
typedef scanner<iterator_t> scanner_t;

class QB_parser;

namespace
{
    unsigned option_base = 0;

    SubFuncP in_sub = 0;
    bool in_block = false;

    struct CoordType
    {
        bool step;
        exprp x,y;
    };
    CoordType coord_res;

    struct Context
    {
    public:
        exprp exp_result;
        stmtp stm_result;
        TypeP typ_result;
    public:
        Context(): exp_result(0), stm_result(0), typ_result(0)
        {
        }
    };
    Context ctx;
    std::list<Context> saved_contexts;

    void SetExpr(exprp new_res)
    {
        ctx.exp_result = new_res ? new_res->optimized() : new_res;
    }
    void SetStmt(stmtp new_stm)
    {
        if(new_stm && ctx.stm_result)
        {
            stmtp prev_getter = new_stm;
            while(prev_getter->get_prev())
            {
                prev_getter = prev_getter->get_prev();
            }
            prev_getter->set_prev(ctx.stm_result);
        }
        ctx.stm_result = new_stm;
    }
    void SetStmt(stmtp new_stm, stmtp prev)
    {
        if(prev)new_stm->set_prev(prev);
        ctx.stm_result = new_stm;
    }
    void SetType(TypeP new_type)
    {
        ctx.typ_result = new_type;
    }

    /////////////////////////////////

    void DoNumber(scanner_t& scan)
    {
        TypeP numtype;
        switch(scan.GetInfo().numtype)
        {
            case '%': numtype = TypeInt; break;
            case '!': numtype = TypeSng; break;
            default:
            case '#': numtype = TypeDbl; break;
            case '&': numtype = TypeLng; break;
        }
        SetExpr(new expr_number(scan.GetInfo().number, numtype));
    }
    void DoString(scanner_t& scan)
    {
        SetExpr(new expr_string(scan.GetInfo().str));
    }

    void DoNumericLabelRef(scanner_t& scan)
    {
        SetExpr(new expr_label( std::to_string( BasCINT(scan.GetInfo().number) ) ));
    }
    void DoIdLabelRef(scanner_t& scan)
    {
        std::string id = scan.GetInfo().str;
        if(HasTypeSuffix(id)) throw InvalidateAction("Not a label");
        SetExpr(new expr_label( DeclareAsLabelInCurrentScope(id, false) ));
    }
    void DoNumericLabelDef(scanner_t& scan)
    {
        std::string name = std::to_string( BasCINT(scan.GetInfo().number) );
        if(!AnchorLabelHere(name)) throw InvalidateAction("Duplicate label");
        SetStmt(new stmt_label(scan.GetLoc(), name));
    }
    void DoIdLabelDef(scanner_t& scan)
    {
        std::string id = scan.GetInfo().str;
        if(HasTypeSuffix(id)) throw InvalidateAction("Not a label");
        std::string name = DeclareAsLabelInCurrentScope(id, true);
        if(name.empty()) throw InvalidateAction("Duplicate label");
        if(!AnchorLabelHere(name)) throw InvalidateAction("Duplicate label");
        SetStmt(new stmt_label(scan.GetLoc(), name));
    }

    bool Do_Dim(const LexLocation& lexloc,
                const std::string& varname,
                const std::vector<std::pair<exprp, exprp> >& dim_exprs,
                TypeP type,
                bool is_redim,
                bool is_shared,
                bool is_static)
    {
        if(type && HasTypeSuffix(varname))
        {
            throw InvalidateAction("Type suffix not allowed in DIM..AS");
        }

#ifdef BINDING_DEBUG
        fprintf(stderr, "1: %s %s(%u)\n",
            is_redim ? "redim" : "dim",
            varname.c_str(),
            (unsigned)dim_exprs.size());
#endif
        // Scalar:
        if(dim_exprs.empty())
        {
            if(is_redim)
            {
                // REDIM is invalid with scalars
                throw InvalidateAction("Invalid function call");
            }
            VarContext previous_def = FindVariableFromScope(varname, false);
            if(previous_def)
            {
                throw InvalidateAction("Duplicate definition");
            }
            if(type)
                return ExplicitlyDefine(varname, type, is_static, is_shared);

            return ImplicitlyDefineScalar(varname, is_static, is_shared);
        }

        // Array:
        VarContext previous_def = FindVariableFromScope(varname, true);
#ifdef BINDING_DEBUG
        if(previous_def)
            fprintf(stderr, "1b: Array type is %s\n",
                previous_def.info->type->arraytype == TypeInfo::DynamicArray
                    ? "dyn" : "other than dyn");
#endif
        bool dims_are_const = true;
        for(unsigned a=0; a<dim_exprs.size(); ++a)
            if(!dim_exprs[a].first->is_const() || !dim_exprs[a].first->is_const())
            {
                dims_are_const = false;
                break;
            }

        bool create_dynamic = IsMetaDynamic() || !dims_are_const;
        if(is_redim && !previous_def)
        {
            create_dynamic = true;
        }
        if(is_redim && previous_def && previous_def.info->type->arraytype == TypeInfo::DynamicArray)
        {
            create_dynamic = true;
        }

        if(previous_def
        && (!is_redim
         || (previous_def.info->type->arraytype != TypeInfo::DynamicArray
          && previous_def.info->type->arraytype != TypeInfo::ParamArray)
           ))
        {
            throw InvalidateAction("Array already dimensioned");
        }

        is_redim = previous_def;

        /*
           At this point, possibilities are:

                previous_def  is_redim  create_dynamic

                        NONE        NO              NO
                        NONE        NO             YES
                     DYNAMIC       YES             ---

         */

#ifdef BINDING_DEBUG
        fprintf(stderr, "2: %s %s array %s (%s)\n",
            is_redim ? "redim" : "dim",
            create_dynamic ? "dyn":"stc",
            varname.c_str(),
            previous_def ? "previously known" : "previously unknown");
#endif
        if(is_redim)
        {
            /* Check the number of dimensions, and if it matches,
             * generate a REDIM runtime call. */
            if(type)
            {
                if(*type != *previous_def.info->type)
                    throw InvalidateAction("Type mismatch");
            }
            type = previous_def.info->type;

            if(type->arraytype == TypeInfo::ParamArray)
            {
                // always run REDIM, viability is checked runtime
                create_dynamic = true;
            }
            else
            {
                if(type->dimensions.size() != dim_exprs.size())
                {
                    throw InvalidateAction("Array dimensions are different");
                }
            }
        }
        else
        {
            /* Create the variable first. */
            if(!type) type = CreateTypeTemplate(varname);

            // Default array dimensions
            type->dimensions.resize(dim_exprs.size(),
                                    std::make_pair((long)option_base,
                                                   (unsigned)(10+(1-option_base))));
            type->arraytype = create_dynamic ? TypeInfo::DynamicArray : TypeInfo::StaticArray;

            //if(!create_dynamic)
            {
                /* Create the specific dimensions. */
                for(unsigned a=0; a<dim_exprs.size(); ++a)
                {
                    expr_number* begin = dynamic_cast<expr_number*> (&*dim_exprs[a].first->optimized());
                    expr_number* end   = dynamic_cast<expr_number*> (&*dim_exprs[a].second->optimized());

                    if(!begin || !end) continue;

                    long long int min = begin->get_i(), count = end->get_i() - min + 1;
                    if(count < 0) throw InvalidateAction("Invalid function call");

                    type->dimensions[a].first  = min;
                    type->dimensions[a].second = count;
                }
            }

            if(!ExplicitlyDefine(varname, type, is_static, is_shared))
            {
                throw InvalidateAction("Duplicate definition");
            }
            /* Then, if dynamic, generate a REDIM function call. */
        }

        if(create_dynamic)
        {
            expr_var* e = new expr_var(varname);
            e->set_type(type);
            e->setwholearray();
            SetStmt(new stmt_redim(lexloc, e, dim_exprs));
            //fprintf(stderr, "redim stmt created\n");
        }

        return true;
    }

    struct ExprPrioItem
    {
        exprp    exp;
        QBtoken  tok;
        unsigned prio;
    };
};

struct Ctx1 /* enter context */
{
    void operator() () const
    {
        saved_contexts.push_back(ctx);
        ctx = Context();
    }
};
struct Ctx0 /* exit context */
{
    void operator() () const
    {
        exprp e = ctx.exp_result;
        stmtp s = ctx.stm_result;
        TypeP t = ctx.typ_result;
        ctx = saved_contexts.back();
        saved_contexts.pop_back();

        /* If the subcontext produced an expression or a statement,
         * import them to the calling context
         */
        if(e)SetExpr(e);
        if(s)SetStmt(s);
        if(t)SetType(t);

#ifdef PARSER_DEBUG
        dprintf(dp_rulecomego, "> pop ctx, e=%s stmt=%s[prev=%s] type=%s\n",
            ctx.exp_result ? typenameof(*ctx.exp_result).c_str() : "NULL",
            ctx.stm_result ? typenameof(*ctx.stm_result).c_str() : "NULL",
            ctx.stm_result && ctx.stm_result->get_prev() ?
                    typenameof(*ctx.stm_result->get_prev()).c_str() : "NULL",
            ctx.typ_result ? typenameof(*ctx.typ_result).c_str() : "NULL");
#endif
    }

};

template<typename RetType>
static bool ParseLibItem(QB_parser& parser, scanner_t& scan, QBtoken tok,
    const char* defs, RetType*& result,
    std::string& matched_pattern);

template<typename StmtType>
static bool ParseLibCommand(QB_parser& parser, scanner_t& scan, QBtoken tok,
                            const char* defs)
{
    StmtType* result = 0;
    std::string matched_pattern;
    if(!ParseLibItem(parser, scan, tok, defs, result, matched_pattern)) return false;
    result->set_pattern(matched_pattern);
    SetStmt(result);
    return true;
}

template<typename ExprType>
static bool ParseLibFunction(QB_parser& parser, scanner_t& scan, QBtoken tok,
                             const char* defs)
{
    ExprType* result = 0;
    std::string matched_pattern;
    if(!ParseLibItem(parser, scan, tok, defs, result, matched_pattern)) return false;
    result->set_pattern(matched_pattern);
    SetExpr(result);
    return true;
}


typedef rule<scanner_t, Ctx1, Ctx0> rule_t;

class QB_parser: public parser<QB_parser>
{
    #define tk(x) ch_p((QBtoken)(x))
    #define tq(x) ch_p((QBtoken)(x)).expected()

    #define num_expr           expr
    #define string_expr        expr
    #define string_or_num_expr expr

    #define LambdaParseBody(x)  ((x).parse(scan))

    #define TypeCheckAny(expr) {}

    #define TypeCheckNum(expr) \
        if(expr->is_num().is_false()) \
           throw InvalidateAction("Type mismatch");

    #define TypeCheckString(expr) \
        if(expr->is_string().is_false()) \
           throw InvalidateAction("Type mismatch");

    #define LambdaExpectToken(tkn, ifnot) \
        do { if(!tq(tkn).parse(scan)) { ifnot; }; } while(0)

    #define LambdaExpectSubRule(rule, ifnot) \
        do { if(!(rule).expected().parse(scan)) { ifnot; }; } while(0)

    #define LambdaExpectAnyExpr(expr, ifnot) \
     do { if(!expr.expected().parse(scan)) { ifnot; }; \
        } while(0)

    #define LambdaExpectNumExpr(expr, ifnot) \
     do { LambdaExpectAnyExpr(expr, ifnot); \
          TypeCheckNum(ctx.exp_result); \
        } while(0)

    #define LambdaExpectStringExpr(expr, ifnot) \
     do { LambdaExpectAnyExpr(expr, ifnot); \
          TypeCheckString(ctx.exp_result); \
        } while(0)

public:
    rule_t program;


    rule_t separator = switch_p [ case_p<COLONCHAR> (eps_p),
                                  case_p<CRLF> (eps_p)
                                ];


    rule_t type = LambdaTokParser(
        [] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                case INTEGER: SetType(new TypeInfo(*TypeInt)); return true;
                case SINGLE:  SetType(new TypeInfo(*TypeSng)); return true;
                case DOUBLE:  SetType(new TypeInfo(*TypeDbl)); return true;
                case LONG:    SetType(new TypeInfo(*TypeLng)); return true;
                case qb64BYTE: SetType(new TypeInfo(*TypeByte)); return true;
                case qb64UNSIGNEDBYTE:    SetType(new TypeInfo(*TypeUbyte)); return true;
                case qb64UNSIGNEDINTEGER: SetType(new TypeInfo(*TypeUint));  return true;
                case qb64UNSIGNEDLONG:    SetType(new TypeInfo(*TypeUlong)); return true;
                case STRING:
                {
                    TypeP t = new TypeInfo(TypeType::String);
                    if((tk('*').parse(scan)))
                    {
                        tq(NUMBER).parse(scan);
                        t->fixlength = BasCINT(scan.GetInfo().number);
                    }
                    SetType(t);
                    return true;
                }
                case IDENTIFIER:
                {
                    TypeP t = new TypeInfo(TypeType::UserType);
                    const std::string utypename = scan.GetInfo().str;
                    if(HasTypeSuffix(utypename))
                        throw InvalidateAction("A type name cannot have a type suffix");

                    t->members   = FindUserType(utypename);

                    if(!t->members)
                        throw InvalidateAction("Undefined type");

                    t->utypename = BasUCASE(StripTypeSuffix(utypename));

                    SetType(t);
                    return true;
                }
                default: break;
            }
            return false;
        }, "type").expected();


    rule_t anylabeltarget;
    rule_t plain_coordinate;
    rule_t step_coordinate;

    rule_t longstatement;
    rule_t shortstatement;
    rule_t colonchars_shortstatements_and_colonchar;
    rule_t shortif;
    rule_t casematch, simplestatement;

    rule_t expr_meat;
    rule_t var_target;
    rule_t autofilenum;
    rule_t ambiguous_id;

    rule_t expr, func_param;

    template<typename Lambda1,typename Lambda2>
    bool ParseVarDef(scanner_t& scan, const Lambda1& handle_const, const Lambda2& handle_sub)
    {
        if(!ambiguous_id.parse(scan))
        {
            // fail quietly, because this rule is an alternative
            return false;
        }

        /* If it has indexes, it can be ARRAY/FUNCTION/DEF name.
         * If it has no indexes, it can be CONST/SCALAR/FUNCTIONNAME.
         * If it has members, it was an user type.
         */
        expr_var* var = dynamic_cast<expr_var*> (&*ctx.exp_result);
        if(!var) throw InvalidateAction("Invalid result from ambiguous_id");

        const std::string name = var->getname();

        bool is_static = in_sub && in_sub->is_static;

        bool can_be_scalar = true;
        bool can_be_array  = true;
        bool can_be_funct  = true;
        bool can_be_const  = true;

        if(var->hasmember())
        {
            // If it has '.', it cannot be a CONST, a FUNCTION or a DEF.
            // It can still be an array or a scalar.
            can_be_const  = false;
            can_be_funct  = false;
        }

        if(var->isarray())
        {
            // If it has '(', it cannot be a CONST or a scalar.
            // It can still be an array, a FUNCTION call, or a DEF call.
            can_be_const  = false;
            can_be_scalar = false;
            for(unsigned b=var->numindexes(), a=0; a<b; ++a)
            {
                if(expr_var* v = dynamic_cast<expr_var*>
                                           (&* var->getindex(a)))
                    if(v->iswholearray())
                    {
                        // "a()" is only valid in a function call.
                        // It cannot be an array index.
                        can_be_array = false;
                    }
            }
        }
        else
        {
            //can_be_array = false; -- it can still be an array!
            // TODO: Explain how. Maybe the GET statement?
        }

        /* Create the relevant type of expression */

#ifdef BINDING_DEBUG
        dprintf(dp_declare, "%s may be: %s %s %s %s\n",
            name.c_str(),
            can_be_scalar ? "scalar" : "",
            can_be_array  ? "array" : "",
            can_be_funct  ? "funct" : "",
            can_be_const  ? "const" : "");
#endif
        if(can_be_const)
        {
            exprp c = FindConstant(name);
            if(c) return handle_const(var, c);
        }
        if(can_be_funct)
        {
            SubFuncP s = FindSubFunc(name);
            if(s) return handle_sub(var, s, var->isarray());
        }

        TypeP got_var;
        if(can_be_scalar)
        {
            VarContext c = FindVariableFromScope(name, false);
            if(c.info && c.info->type)
            {
                got_var = c.info->type;
            }
            else
            {
                // Implicitly define. Can however only implicitly define standard types.
                if(var->hasmember())
                    throw InvalidateAction("Not defined as an user type (missing DIM?): " + name);
                if(!var->isarray())
                {
                    got_var = ImplicitlyDefineScalar(name, is_static);
#ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Tried to implicitly define scalar %s - %s\n",
                        name.c_str(), got_var?"success":"fail");
#endif
                }
            }
        }
        if(can_be_array && !got_var)
        {
            VarContext c = FindVariableFromScope(name, true);
            if(c.info && c.info->type)
            {
                got_var = c.info->type;
            }
            else
            {
                // Implicitly define. Can however only implicitly define standard types.
                if(var->hasmember())
                    throw InvalidateAction("Not defined as an user type (missing DIM?): " + name);
                if(var->isarray())
                {
                    got_var = ImplicitlyDefineArray(name,
                                                    var->numindexes(),
                                                    option_base, 10,
                                                    is_static);
#ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Tried to implicitly define array %s - %s\n",
                        name.c_str(), got_var?"success":"fail");
#endif
                }
            }
        }
        if(got_var)
        {
            TypeP pivot_found           = new TypeInfo(*got_var);
            autoptr<expr_var> pivot_var = var;

        DeeperCheck:
#ifdef BINDING_DEBUG
            dprintf(dp_declare, "DeeperCheck: pivot(%s)=%s\n",
                name.c_str(), pivot_found->Dump().c_str());
#endif
            /* Check the members and number of array dimensions for match */
            if(pivot_found->arraytype != TypeInfo::ParamArray)
            {
                if(pivot_var->numindexes() != pivot_found->dimensions.size())
                    throw InvalidateAction("Array dimensions do not match");

                // addressing an element, not the array
                pivot_found->dimensions.clear();
                pivot_found->arraytype = TypeInfo::NotArray;
            }
            else
            {
                // Parameter arrays are more complicated, because their
                // dimensions are unknown
                // Assume the dimensions were matching. Check runtime!
                pivot_found->dimensions.clear();
                pivot_found->arraytype = TypeInfo::NotArray;
            }

            if(pivot_var->hasmember())
            {
                if(!pivot_found->members)
                    throw InvalidateAction("Cannot pass a struct as rvalue");

                autoptr<expr_var> mem = pivot_var->getmember();
                const std::string mem_name = mem->getname();

                TypeP memtype = pivot_found->members->GetMember(mem_name);
                if(!memtype) throw InvalidateAction("No such member variable");
                mem->set_type(memtype);

                pivot_found = memtype;
                pivot_var   = mem;
                goto DeeperCheck;
            }

            var->set_type(pivot_found);
        }
        else
        {
            throw InvalidateAction("Duplicate definition?");
        }

        return true;
    }

public:
    QB_parser()
    {
    program = LambdaParser([&] (scanner_t& scan) -> bool
    {
        /* THE MAIN RULE */

        /* Parse long statements */
        while(!scan.eof() && longstatement.parse(scan));

        /* Check what was left over. */
        while(!scan.eof())
        {
            QBtoken tok = scan.get();
            switch(tok)
            {
                case COLONCHAR:
                case CRLF: break;

                case ENDIF:        throw InvalidateAction("END IF without IF");
                case ENDDEF:       throw InvalidateAction("END DEF without DEF");
                case ENDSUB:       throw InvalidateAction("END SUB without SUB");
                case ENDFUNCTION:  throw InvalidateAction("END FUNCTION without FUNCTION");
                case ENDSELECT:    throw InvalidateAction("END SELECT without SELECT");
                case LOOP:         throw InvalidateAction("LOOP without DO or WEND without WHILE");
                case LOOPWHILE:    throw InvalidateAction("LOOP without DO");
                case LOOPUNTIL:    throw InvalidateAction("LOOP without DO");
                case ENDTYPE:      throw InvalidateAction("END TYPE without TYPE");
                case NEXT:         throw InvalidateAction("NEXT without FOR");
                default: throw InvalidateAction("Unexpected " + GetTokenName(tok));
            }
        }
        return true;
    });

    /**** LVAL */
    var_target = LambdaParser(
        [this] (scanner_t& scan) -> bool
        {
            /* It cannot be a CONST (parse error)
             * It cannot be a FUNCTIONNAME (fail rule without error)
             *   However, it can be the name of current function,
             *   or the name of current DEF.
             */
            bool res = this->ParseVarDef(scan,
                [&] (expr_var*, exprp) -> bool
                { // const
                    throw InvalidateAction("Invalid use of a CONST");
                },

                [&] (expr_var* var, SubFuncP s, bool params) -> bool
                { // sub
                    if(!s->returntype)
                    {
#ifdef BINDING_DEBUG
                        //dprintf(dp_declare, "Sub call?");
#endif
                        return false; // It may be a sub call instead!
                    }

                    const std::string name = var->getname();
                    const std::string basename = BasUCASE(StripTypeSuffix(name));

                    if(params || basename != GetCurrentSubBaseName())
                    {
                        throw InvalidateAction("Invalid function call syntax");
                    }

                    SetExpr(new expr_returnvalueholder(name));
#ifdef BINDING_DEBUG
                    //fprintf(stderr, "retval \n");
#endif
                    return true;
                }
            );
            if(!res) return false;


            return true;
        }, "lvalue");

    autofilenum = LambdaParser(
        [this] (scanner_t& scan) -> bool
        {
            if(tk(FILENUM).parse(scan))
                { DoNumber(scan); return true; }
            LambdaExpectNumExpr(expr, return false);
            return true;
        }, "file number");

    anylabeltarget = LambdaTokParser(
        [this] (QBtoken tok, scanner_t& scan) -> bool
        {
            if(tok == NUMBER)
                { DoNumericLabelRef(scan); return true; }
            if(tok == IDENTIFIER)
                { DoIdLabelRef(scan); return true; }
            return false;
        });

    plain_coordinate = LambdaParser(
        [this] (scanner_t& scan) -> bool
        {
            if(!tk('(').parse(scan)) return false;
            if(!num_expr.parse(scan)) return false;
            coord_res.x = ctx.exp_result;
            if(!tk(',').parse(scan)) return false;
            if(!num_expr.parse(scan)) return false;
            coord_res.y = ctx.exp_result;
            if(!tk(')').parse(scan)) return false;
            return true;
        }, "coordinate");

    step_coordinate = LambdaParser(
        [this] (scanner_t& scan) -> bool
        {
            coord_res.step = tk(STEP).parse(scan);
            return plain_coordinate.parse(scan);
        }, "step coordinate");

    longstatement =
        /* A label (that does not end in a newline/terminator) */
        LambdaTokParser(
        [&] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                /* a label statement */
                /* must be checked before simplestatement */
                case NUMBER:
                    DoNumericLabelDef(scan); return true;
                case IDENTIFIER:
                    // Labels should only be recognized in the beginning of line
                    // This conflicts with parameterless SUB calls
                    if(!tk(COLONCHAR).parse(scan)) break;
                    DoIdLabelDef(scan); return true;

                // filler
                case COLONCHAR: return true;
                case CRLF: return true;
                default:break;
            }
            return false;
        })
        /* Or, a statement (that must end in a newline/terminator) */
    | ( LambdaTokParser(
        [this] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                case IF:
                {
                    stmtp prev = ctx.stm_result;
                    const block_setting<bool> block_marker(in_block, true);

                    LambdaExpectNumExpr((expr), break);

                    autoptr<stmt_if> if_stmt = new stmt_if(scan.GetLoc());
                    if_stmt->condition = ctx.exp_result;

                    /* Use tk instead of tq, to enable shortif without problems */
                    if(!tk(THEN).parse(scan)) break;
                    if(!tk(CRLF).parse(scan)) break;

                    SetStmt(0);
                    while((longstatement).parse(scan));

                    if_stmt->thenp = ctx.stm_result;

                    while(tk(ELSEIF).parse(scan))
                    {
                        LambdaExpectNumExpr((expr), break);
                        exprp ex = ctx.exp_result;

                        tq(THEN).parse(scan);

                        SetStmt(0);
                        while((longstatement).parse(scan));
                        stmtp st = ctx.stm_result;

                        if_stmt->elseifp.push_back(std::make_pair(ex, st));
                    }
                    if(tk(ELSE).parse(scan))
                    {
                        SetStmt(0);
                        while((longstatement).parse(scan));
                        if_stmt->elsep = ctx.stm_result;
                    }
                    SetStmt(&*if_stmt, prev);
                    return tq(ENDIF).parse(scan);
                }

                case SELECT:
                {
                    stmtp prev = ctx.stm_result;
                    const block_setting<bool> block_marker(in_block, true);

                    LambdaExpectToken(CASE, break);

                    LambdaExpectAnyExpr((expr), break);

                    autoptr<stmt_select> select_stmt = new stmt_select(scan.GetLoc());
                    select_stmt->ref = ctx.exp_result;

                    LambdaExpectSubRule((separator), break);
                    for(;;)
                    {
                        while(separator.parse(scan)) {}
                        if(!tk(CASE).parse(scan)) break;

                        LambdaExpectSubRule((casematch), return false);
                        exprp cond = ctx.exp_result;
                        LambdaExpectSubRule((separator), break);

                        SetStmt(0);
                        while((longstatement).parse(scan));

                        stmtp st = ctx.stm_result;
                        select_stmt->cases.push_back(std::make_pair(cond, st));
                    }
                    SetStmt(&*select_stmt, prev);
                    return tq(ENDSELECT).parse(scan);
                };


                case FOR:
                {
                    stmtp prev = ctx.stm_result;
                    const block_setting<bool> block_marker(in_block, true);

                    if(!(this->ParseVarDef)(scan,
                        [&] (expr_var*, exprp) -> bool
                        { // const
                            throw InvalidateAction("Invalid use of a CONST");
                        },
                        [&] (expr_var*, SubFuncP, bool) -> bool
                        { // sub
                            throw InvalidateAction("Invalid use of a SUB name");
                        }
                    )) break;
                    expr_var* vardef = dynamic_cast<expr_var*>(&*ctx.exp_result);
                    const std::string& loopvar = vardef->getname();
                    TypeP vartype = vardef->GetType();
                    if(!vartype) throw InvalidateAction("Duplicate definition");
                    if(!vartype->is_num()) throw InvalidateAction("Type mismatch");
                    if(vardef->isarray()) throw InvalidateAction("FOR loop variable cannot be an array member");

                    LambdaExpectToken('=', break);

                    autoptr<stmt_for> for_stmt = new stmt_for(scan.GetLoc());
                    for_stmt->var = vardef;

                    LambdaExpectNumExpr((expr),  break );
                    for_stmt->begin = ctx.exp_result;

                    LambdaExpectToken(TO, break);
                    LambdaExpectNumExpr((expr),  break );
                    for_stmt->end = ctx.exp_result;
                    if(tk(STEP).parse(scan))
                    {
                        LambdaExpectNumExpr((expr),  break );
                        for_stmt->step = ctx.exp_result;
                    }
                    LambdaExpectSubRule((separator), break);

                    SetStmt(0);
                    while((longstatement).parse(scan));

                    for_stmt->bodyp = ctx.stm_result;

                    LambdaExpectToken(NEXT, break);

                    if(tk(IDENTIFIER).parse(scan))
                    {
                        std::string gotvar = scan.GetInfo().str;
                        if(gotvar != loopvar)
                            throw InvalidateAction
                                ("NEXT without FOR (expected "+loopvar+", got "+gotvar+")");
                    }

                    SetStmt(&*for_stmt, prev);

                    return true;
                }

                 /* do while ... loop
                    do until ... loop
                    do ... loop
                    do ... loop while
                    do ... loop until */
                case DO:
                case DOWHILE:
                case DOUNTIL:
                {
                    stmtp prev = ctx.stm_result;
                    const block_setting<bool> block_marker(in_block, true);

                    autoptr<stmt_doloop> doloop_stmt = new stmt_doloop(scan.GetLoc());
                    if(tok != DO)
                    {
                        LambdaExpectNumExpr((expr), );
                        if(tok == DOUNTIL)
                            ctx.exp_result = new expr_unary(NOT, ctx.exp_result);
                        doloop_stmt->precondition = ctx.exp_result;
                    }
                    LambdaExpectSubRule((separator), break);

                    SetStmt(0);
                    while((longstatement).parse(scan));

                    doloop_stmt->bodyp = ctx.stm_result;
                    QBtoken tok2 = scan.get();
                    switch(tok2)
                    {
                        case LOOP: break;
                        case LOOPWHILE:
                        case LOOPUNTIL:
                            if(tok == DO)
                            {
                                LambdaExpectNumExpr((expr), break);
                                if(tok2 == LOOPUNTIL)
                                    ctx.exp_result = new expr_unary(NOT, ctx.exp_result);
                                doloop_stmt->postcondition = ctx.exp_result;
                                break;
                            }
                        default: return false;
                    }
                    SetStmt(&*doloop_stmt, prev);
                    return true;
                }

                /* Note: SUBs cannot exist in SUBs. */
                case SUB:
                case FUNCTION:
                case DEF:
                case DECLARESUB:
                case DECLAREFUNCTION:
                {
                    bool is_declare = tok == DECLARESUB || tok == DECLAREFUNCTION;
                    stmtp prev = ctx.stm_result;

#ifdef BINDING_DEBUG
                    dprintf(dp_declare, "Declare sub/func/def...\n");
#endif
                    if(in_sub || in_block)
                        throw InvalidateAction("SUB/FUNCTION/DEF/DECLARE not allowed in control statements");

                    QBtoken endtoken = tok==SUB ? ENDSUB : ENDFUNCTION;

                    if(tok == DEF)
                    {
                        LambdaExpectToken(FNIDENTIFIER, break);
                        endtoken = ENDDEF;
                    }
                    else
                    {
                        LambdaExpectToken(IDENTIFIER, break);
                    }
                    std::string subname = scan.GetInfo().str;
                    std::vector<SubRoutine::ParamSpec> params;
                    bool is_cdecl = false;
                    bool is_static = false;
                    std::string alias;

                    if((tok == SUB || tok == DECLARESUB) && HasTypeSuffix(subname))
                        throw InvalidateAction("Invalid SUB name");

                    if(tk('(').parse(scan) && !tk(')').parse(scan)) /* optional */
                    {
                        do {
                            SubRoutine::ParamSpec param;

                            if(tk(BYVAL).parse(scan)) { param.SetByval(); }
                            else if(tk(SEG).parse(scan)) { param.SetSeg(); }

                            LambdaExpectToken(IDENTIFIER, return false);
                            param.name     = scan.GetInfo().str;
                            param.is_array = false;
                            param.was_modified = true; // assume worst
                            if(tk('(').parse(scan))
                            {
                                LambdaExpectToken(')', break);
                                param.is_array = true;
                            }

                            param.type = CreateTypeTemplate(param.name);
                            if(tk(AS).parse(scan))
                            {
                                if(HasTypeSuffix(param.name))
                                {
                                    throw InvalidateAction("Invalid AS together with type suffix");
                                }
                                if(tk(ANY).parse(scan))
                                {
                                    if(!is_declare || !param.is_array)
                                        throw InvalidateAction("AS ANY is only valid with arrays in DECLARE");
                                    param.type = 0;
                                }
                                else
                                {
                                    LambdaExpectSubRule((type), return false);
                                    param.type = ctx.typ_result;
                                }
                            }
                            if(param.type && param.is_array)
                            {
                                param.type->arraytype = TypeInfo::ParamArray;
                            }
                            params.push_back(param);
                        } while(tk(',').parse(scan));
                        LambdaExpectToken(')', break);
                    }
                    else
                    {
                        if(is_declare)
                            throw InvalidateAction("Expected: '('");
                    }

                    if(!is_declare && tk(STATIC).parse(scan)) // optional
                    {
                        if(tok == DEF)
                            throw InvalidateAction("DEFs cannot be 'STATIC'");
                        else
                            is_static = true;
                    }


                    if(is_declare)
                    {
                        if(tk(CDECL).parse(scan)) { is_cdecl = true;  }
                        if(tk(ALIAS).parse(scan))
                        {
                            LambdaExpectToken(STRINGCONST, break);
                            alias = scan.GetInfo().str;
                        }
                    }

                    iterator_t it = scan.pos();
                    if(!tk(CRLF).parse(scan))
                    {
                        if(tok == DEF) break; // no harm, try an oneliner format later!
                        // otherwise (SUB/FUNCTION):
                        throw InvalidateAction("Expected: CRLF");
                    }
                    else
                    {
                        if(tok == DECLARESUB || tok == DECLAREFUNCTION)
                        {
                            // Revert the CRLF so that subrule "separator"
                            // will match properly.
                            scan.revert(it);
                        }
                    }

                    TypeP returntype;
                    if(tok == FUNCTION
                    || tok == DECLAREFUNCTION
                    || tok == DEF)
                        returntype = CreateTypeTemplate(subname);

                    SubFuncP sub = DefineSubFunc(is_declare, subname,
                                                 returntype,
                                                 params);
                    if(!sub) throw InvalidateAction("Duplicate definition");
                    if(is_cdecl) sub->callconvention = 1;
                    if(!alias.empty()) sub->linkername = alias;
                    sub->is_static = is_static;

                    if(!is_declare)
                    {
                        /* For DEFs, ALL variables are SHARED even if not declared SHARED */
                        EnterSubFuncScope(subname, tok == DEF);

                        const block_setting<SubFuncP> block_marker(in_sub, sub);

                        SetStmt(0);
                        while((longstatement).parse(scan));

                        ExitSubFuncScope(subname, ctx.stm_result);

                        if(!tq(endtoken).parse(scan))
                        {
                            // It gave a warning. Check if there's another type
                            // of end-token instead.
                            if(!(tk(ENDSUB) | tk(ENDFUNCTION) | tk(ENDDEF)).parse(scan))
                            {
                                // FIXME: This doesn't work.
                                break;
                            }
                        }

                        SetStmt(prev, (stmt*)0);
                    }
                    return true;
                }

                /* Note: TYPE cannot exist in SUBs.
                 * However, Quickbasic allows them inside control structures.
                 * But they are unconditional.
                 */
                case TYPE:
                {
                    if(in_sub)
                        throw InvalidateAction("TYPE not allowed in SUB/FUNCTION/DEF");

                    LambdaExpectToken(IDENTIFIER, break);
                    const std::string utypename = scan.GetInfo().str;

                    if(HasTypeSuffix(utypename))
                        throw InvalidateAction("TYPE name cannot have a type suffix");

                    UtypeP utype = new UserType;

                    LambdaExpectToken(CRLF, break);
                    while(!scan.eof())
                    {
                        // empty line is ok in typedef
                        if(tk(CRLF).parse(scan)) continue;

                        if(!tk(IDENTIFIER).parse(scan)) break;

                        while(!scan.eof())
                        {
                            const std::string membername = scan.GetInfo().str;
                            if(HasTypeSuffix(membername))
                                throw InvalidateAction("TYPE member cannot have a type suffix");

                            LambdaExpectToken(AS, return false);
                            LambdaExpectSubRule((type), return false);

                            if(!utype->DeclareMember(membername, ctx.typ_result))
                                throw InvalidateAction("Duplicate definition");

                            if(!tk(',').parse(scan)) break;

                            if(!tq(IDENTIFIER).parse(scan)) break;
                        }
                        LambdaExpectToken(CRLF, break);
                    }
                    LambdaExpectToken(ENDTYPE, break);

                    if(!DefineUserType(utypename, utype))
                    {
                        throw InvalidateAction("Duplicate definition");
                    }

                    return true;
                }

                default: break;
            }
            return false;
        })
      | simplestatement
      | shortif
     ) >> LambdaParser([this] (scanner_t& scan) -> bool
          {
              if(scan.eof()) throw InvalidateAction("File must end in a newline");
              LambdaExpectSubRule((separator), return false);
              return true;
          });

    shortif = LambdaTokParser(
        [this] (QBtoken tok, scanner_t& scan) -> bool
        {
            if(tok != IF) return false;

            stmtp prev = ctx.stm_result;

            const block_setting<bool> block_marker(in_block, true);

            LambdaExpectNumExpr(expr, return false);

            autoptr<stmt_if> if_stmt = new stmt_if(scan.GetLoc());
            if_stmt->condition = ctx.exp_result;

            LambdaExpectToken(THEN, return false);

            bool got_else = false;
            bool got_stmt = false;
            for(;;)
            {
                SetStmt(0);
                if(tk(NUMBER).parse(scan))
                {
                    /* Translate into GOTO */
                    DoNumericLabelRef(scan);
                    SetStmt(new stmt_labelref(scan.GetLoc(), GOTO, ctx.exp_result));
                    got_stmt = true;

                    while(tk(COLONCHAR).parse(scan));
                }
                else
                {
                    bool stmt_ok = true;
                    while(!scan.eof())
                    {
                        if(tk(COLONCHAR).parse(scan))
                        {
                            got_stmt = true;
                            stmt_ok = true;
                            continue;
                        }
                        if(!stmt_ok) break;
                        if(shortstatement.parse(scan))
                        {
                            got_stmt = true;
                            stmt_ok = false;
                            continue;
                        }
                        if(shortif.parse(scan))
                        {
                            got_stmt = true;
                            break;
                        }
                        break;
                    }
                }

                if(got_else)
                    if_stmt->elsep = ctx.stm_result;
                else
                    if_stmt->thenp = ctx.stm_result;

                if(got_else) break;
                if(!tk(ELSE).parse(scan)) break;
                got_else = true;
            }
            SetStmt(&*if_stmt, prev);
            return got_stmt || got_else;
        }, "IF clause");

    colonchars_shortstatements_and_colonchar =
        LambdaParser(
        [this] (scanner_t& scan) -> bool
        {
            SetStmt(0);
            LambdaExpectToken(COLONCHAR, return false);
            bool got_colonchar = true;
            iterator_t it = scan.pos();
            while(!scan.eof())
            {
                if(tk(COLONCHAR).parse(scan))
                {
                    it = scan.pos();
                    got_colonchar = true;
                    continue;
                }
                if(!got_colonchar) break;
                if(!shortstatement.parse(scan)) break;
                got_colonchar=false;
            }
            if(got_colonchar) return true;
            scan.revert(it);
            return false;
        }, "statements");

    /* shortstatement does not include trailing COLONCHARs. */
    shortstatement =
        /* statements that can be included between online THEN and ELSE */
        /* Does not include shortif. */
        simplestatement
    |   LambdaTokParser(
        [&] (QBtoken tok, scanner_t& scan) -> bool
        {
            switch(tok)
            {
                case FOR:
                {
                    stmtp prev = ctx.stm_result;

                    const block_setting<bool> block_marker(in_block, true);

                    if(!(this->ParseVarDef)(scan,
                        [&] (expr_var*, exprp) -> bool
                        { // const
                            throw InvalidateAction("Invalid use of a CONST");
                        },
                        [&] (expr_var*, SubFuncP, bool) -> bool
                        { // sub
                            throw InvalidateAction("Invalid use of a SUB name");
                        }
                    )) break;
                    expr_var* vardef = dynamic_cast<expr_var*>(&*ctx.exp_result);
                    const std::string& loopvar = vardef->getname();
                    TypeP vartype = vardef->GetType();
                    if(!vartype) throw InvalidateAction("Duplicate definition");
                    if(!vartype->is_num()) throw InvalidateAction("Type mismatch");
                    if(vardef->isarray()) throw InvalidateAction("FOR loop variable cannot be an array member");

                    LambdaExpectToken('=', break);

                    autoptr<stmt_for> for_stmt = new stmt_for(scan.GetLoc());
                    for_stmt->var = vardef;

                    LambdaExpectNumExpr((expr),  break );
                    for_stmt->begin = ctx.exp_result;

                    LambdaExpectToken(TO, break);
                    LambdaExpectNumExpr((expr),  break );
                    for_stmt->end = ctx.exp_result;
                    if(tk(STEP).parse(scan))
                    {
                        LambdaExpectNumExpr((expr),  break );
                        for_stmt->step = ctx.exp_result;
                    }

                    LambdaExpectSubRule((colonchars_shortstatements_and_colonchar), break);
                    for_stmt->bodyp = ctx.stm_result;

                    LambdaExpectToken(NEXT, break);

                    if(tk(IDENTIFIER).parse(scan))
                    {
                        std::string gotvar = scan.GetInfo().str;
                        if(gotvar != loopvar)
                            throw InvalidateAction
                                ("NEXT without FOR (expected "+loopvar+", got "+gotvar+")");
                    }

                    SetStmt(&*for_stmt, prev);

                    return true;
                }
                case SELECT:
                {
                    stmtp prev = ctx.stm_result;

                    const block_setting<bool> block_marker(in_block, true);

                    LambdaExpectToken(CASE, break);
                    LambdaExpectAnyExpr((expr), break);

                    autoptr<stmt_select> select_stmt = new stmt_select(scan.GetLoc());
                    select_stmt->ref = ctx.exp_result;

                    LambdaExpectToken(COLONCHAR, break);
                    while(tk(COLONCHAR).parse(scan));

                    while(tk(CASE).parse(scan))
                    {
                        LambdaExpectSubRule((casematch), return false);
                        exprp cond = ctx.exp_result;

                        LambdaExpectSubRule((colonchars_shortstatements_and_colonchar), return false);

                        stmtp st = ctx.stm_result;
                        select_stmt->cases.push_back(std::make_pair(cond, st));
                    }
                    SetStmt(&*select_stmt, prev);
                    return tq(ENDSELECT).parse(scan);
                };

                 /* do while ... loop
                    do until ... loop
                    do ... loop
                    do ... loop while
                    do ... loop until */
                case DO:
                case DOWHILE:
                case DOUNTIL:
                {
                    stmtp prev = ctx.stm_result;

                    const block_setting<bool> block_marker(in_block, true);

                    autoptr<stmt_doloop> doloop_stmt = new stmt_doloop(scan.GetLoc());
                    if(tok != DO)
                    {
                        LambdaExpectNumExpr((expr), );
                        if(tok == DOUNTIL)
                            ctx.exp_result = new expr_unary(NOT, ctx.exp_result);
                        doloop_stmt->precondition = ctx.exp_result;
                    }
                    LambdaExpectSubRule((colonchars_shortstatements_and_colonchar), break);
                    doloop_stmt->bodyp = ctx.stm_result;
                    QBtoken tok2 = scan.get();
                    switch(tok2)
                    {
                        case LOOP: break;
                        case LOOPWHILE:
                        case LOOPUNTIL:
                            if(tok == DO)
                            {
                                LambdaExpectNumExpr((expr), break);
                                if(tok2 == LOOPUNTIL)
                                    ctx.exp_result = new expr_unary(NOT, ctx.exp_result);
                                doloop_stmt->postcondition = ctx.exp_result;
                                break;
                            }
                        default: return false;
                    }
                    SetStmt(&*doloop_stmt, prev);
                    return true;
               }
               default: break;
            }
            return false;
        }, "statement");

    casematch = LambdaParser(
        [&] (scanner_t& scan) -> bool
        {
            exprp prev;
            for(;;)
            {
                if(tk(ELSE).parse(scan))
                {
                    SetExpr(new expr_caseelse);
                }
                else if(tk(IS).parse(scan))
                {
                    bool invert = false;
                    QBtoken comptok = QBtoken();
                    for(iterator_t it = scan.pos(); !scan.eof(); it = scan.pos())
                    {
                        QBtoken tok = scan.get();
                        switch(tok)
                        {
                            case NOT: invert=!invert; break;
                            case (QBtoken)'<':
                            case (QBtoken)'=':
                            case (QBtoken)'>':
                            case NOTEQUAL:
                            case LESSOREQUAL:
                            case GREATEROREQUAL:
                                if(comptok) { tok=QBtoken(); break; } // Allow only one
                                comptok = tok;
                                break;
                            default:
                                tok=QBtoken();
                        }
                        if(tok==QBtoken()) { scan.revert(it); break; }
                    }
                    if(comptok==QBtoken()) comptok = (QBtoken)'='; // '=' is optional
                    if(invert)
                        switch(comptok)
                        {
                            case (QBtoken)'<': comptok = GREATEROREQUAL; break;
                            default:
                            case (QBtoken)'=': comptok = NOTEQUAL; break;
                            case (QBtoken)'>': comptok = LESSOREQUAL; break;
                            case NOTEQUAL: comptok = (QBtoken)'='; break;
                            case LESSOREQUAL: comptok = (QBtoken)'>'; break;
                            case GREATEROREQUAL: comptok = (QBtoken)'<'; break;
                        }
                    LambdaExpectAnyExpr((expr), return false);
                    SetExpr(new expr_caseis(comptok, ctx.exp_result));
                }
                else
                {
                    LambdaExpectAnyExpr((expr), return false);
                    exprp e = ctx.exp_result;
                    if(tk(TO).parse(scan))
                    {
                        LambdaExpectAnyExpr((expr), return false);
                        exprp e2 = ctx.exp_result;
                        SetExpr(new expr_caserange(e, e2));
                    }
                    else
                        SetExpr(new expr_caseis((QBtoken)'=', e));
                }
                if(prev)
                    ctx.exp_result = new expr_caseor(prev, ctx.exp_result);
                prev = ctx.exp_result;

                if(!tk(',').parse(scan)) break;
            }
            return true;
        }, "case option list");

    simplestatement =
        LambdaTokParser
        (
            [&] (QBtoken tok, scanner_t& scan) -> bool
            {
                const char* pattern = 0;

                LexLocation lexloc = scan.GetLoc();

                //fprintf(stderr, "Tok=%s\n", GetTokenName(tok).c_str());

                /* LIBSTATEMENTS */
                switch(tok)
                {
                    default:break;
                    got_stmt_libstmt:
                        return (ParseLibCommand<stmt_lib_stmt>)((*this), scan, tok, pattern);
                    got_stmt_labelref:
                        return (ParseLibCommand<stmt_labelref>)((*this), scan, tok, pattern);

                    #define tokcase(s,p) case s: Do(p)
                    #define Do(p) { pattern=p; goto got_stmt_libstmt; }

#include "libdata-stmt-stmt.inc"

                    #undef Do

                    #define Do(p) { pattern=p; goto got_stmt_labelref; }

#include "libdata-stmt-label.inc"

                    #undef Do
                    #undef tokcase

                    /* More complex statements */

                    case LET:
                    {
                        LambdaExpectSubRule((var_target), break);
                        exprp target = ctx.exp_result;
                        LambdaExpectToken('=', break);
                        LambdaExpectSubRule((expr),break);
                        exprp source = ctx.exp_result;
                        SetStmt(new stmt_let(lexloc, target, source));
                        return true;
                    }

                    /* explicit sub call */
                    case CALL:
                    {
                        LambdaExpectToken(IDENTIFIER, break);
                        // TODO: check namespace
                        const std::string subname = scan.GetInfo().str;
                        std::vector<exprp> params;
                        if(tk('(').parse(scan))
                        {
                            do {
                                LambdaExpectSubRule((func_param), return false);
                                params.push_back(ctx.exp_result);
                            } while(tk(',').parse(scan));
                            LambdaExpectToken(')', break);
                        }
                        SetStmt(new stmt_subcall(lexloc, subname, params));
                        return true;
                    }

                    /* single-line DEF. multiline DEF is already caught in longstatement. */
                    case DEF:
                    {
                        if(in_sub || in_block)
                            throw InvalidateAction("SUB/FUNCTION/DEF not allowed in control statements");

                        stmtp prev = ctx.stm_result;

                        LambdaExpectToken(FNIDENTIFIER, break);
                        std::string subname = scan.GetInfo().str;
                        std::vector<SubRoutine::ParamSpec> params;

                        TypeP returntype = CreateTypeTemplate(subname);

                        // This parameter parsing is mostly duplicate from the multi-line
                        // SUB/FUNCTION/DEF parameter parsing, except here we won't handle
                        // some special cases, like array parameters.

                        if(tk('(').parse(scan) && !tk(')').parse(scan)) /* optional */
                        {
                            do {
                                SubRoutine::ParamSpec param;

                                if(tk(BYVAL).parse(scan)) { param.SetByval(); }

                                LambdaExpectToken(IDENTIFIER, return false);
                                param.name     = scan.GetInfo().str;
                                param.is_array = false;
                                param.was_modified = true; // assume worst

                                param.type = CreateTypeTemplate(param.name);
                                if(tk(AS).parse(scan))
                                {
                                    if(HasTypeSuffix(param.name))
                                    {
                                        throw InvalidateAction("Invalid AS together with type suffix");
                                    }
                                    LambdaExpectSubRule((type), return false);
                                    param.type = ctx.typ_result;
                                }
                                params.push_back(param);
                            } while(tk(',').parse(scan));
                            LambdaExpectToken(')', break);
                        }
                        LambdaExpectToken('=', break);

                        SubFuncP sub = DefineSubFunc(false, subname, returntype, params);
                        if(!sub) throw InvalidateAction("Duplicate definition");

                        /* For DEFs, ALL variables are SHARED even if not declared SHARED */
                        EnterSubFuncScope(subname, true);

                        const block_setting<SubFuncP> block_marker(in_sub, sub);

                        exprp target = new expr_returnvalueholder(subname);
                        LambdaExpectSubRule((expr),break);
                        exprp source = ctx.exp_result;
                        SetStmt(new stmt_let(lexloc, target, source));

                        ExitSubFuncScope(subname, ctx.stm_result);

                        SetStmt(prev, (stmt*)0);
                        return true;
                    }

                    /* Note: QuickBasic allows CONSTs in control structures.
                     * But they are unconditional.
                     */
                    case CONST:
                    {
                        do {
                            LambdaExpectToken(IDENTIFIER, break);
                            std::string id = scan.GetInfo().str;
                            LambdaExpectToken('=', break);
                            LambdaExpectSubRule((expr), break);

                            exprp c = ctx.exp_result->optimized();
                            if(!c->is_const())
                            {
                                throw InvalidateAction("Expression is not constant");
                            }
                            int r = DefineConstant(id, c);
                            if(r == -1) throw InvalidateAction("Type mismatch");
                            if(r == 0) throw InvalidateAction("Duplicate definition");
                        } while( tk(',').parse(scan) );
                        return true;
                    }

                    /* TODO: OPTION BASE cannot exist in SUBs.
                     * Note: QuickBasic allows OPTION BASE in control structures.
                     * But they are unconditional. Also, only 1 is allowed.
                     * Attempt to use more will produce compile-time error
                     * "Array already dimensioned".
                     */
                    case OPTION_BASE:
                    {
                        if(in_sub)
                            throw InvalidateAction("OPTION BASE not allowed in SUB/FUNCTION/DEF");

                        LambdaExpectToken(NUMBER, return false);
                        if(scan.GetInfo().number != 0.0
                        && scan.GetInfo().number != 1.0)
                        {
                            throw InvalidateAction("Expected: 0 or 1\n");
                        }
                        option_base = (unsigned)scan.GetInfo().number;
                        return true;
                    }

                    /* TODO: DATA cannot exist in SUBs.
                     * However, Quickbasic allows them inside control structures.
                     * But they are unconditional.
                     */
                    case DATA:
                        if(in_sub)
                            throw InvalidateAction("DATA not allowed in SUB/FUNCTION/DEF");

                        do {
                            if(tk(STRINGCONST).parse(scan))
                            {
                                //fprintf(stderr, "Data(%s)\n", scan.GetInfo().str.c_str());
                                global_data.push_back(scan.GetInfo().str);
                            }
                            else
                            {
                                //fprintf(stderr, "Data(%s)\n", "");
                                global_data.push_back(std::string());
                            }
                        } while( tk(',').parse(scan) );
                        return true;

                    case SHARED:
                        if(!in_sub)
                            throw InvalidateAction("SHARED invalid outside SUB/FUNCTION");

                        do {
                            tq(IDENTIFIER).parse(scan);
                            std::string gotvar   = scan.GetInfo().str;
                            bool        is_array = false;
                            TypeP       type;
                            if(tk('(').parse(scan))
                            {
                                tq(')').parse(scan);
                                is_array = true;
                            }
                            if(tk(AS).parse(scan))
                            {
                                LambdaExpectSubRule((this->type), return false);
                                type = ctx.typ_result;
                            }

                            SubScopeImportSharedVariable(BasUCASE(gotvar), type, is_array);
                        } while( tk(',').parse(scan) );
                        return true;

                    case COMMON:
                        if(in_sub || in_block)
                            throw InvalidateAction("COMMON not allowed in control statements");

                        fprintf(stderr,
                            "Warning: COMMON not supported yet\n");

                        /* TODO: do */
                        return LambdaParseBody(
                       !tk(SHARED)
                     >> if_p(tk('/')) [ tk(IDENTIFIER) >> '/' ]
                     >> (tq(IDENTIFIER) % ',')
                     >> tq(CRLF)
                    );

                    /* data commands */
                    case READ:
                        do {
                            LambdaExpectSubRule((var_target), return false);

                            SetStmt(new stmt_read(lexloc, ctx.exp_result));
                        } while(tk(',').parse(scan));
                        return true;

                    case DIM:
                    case REDIM:
                    case STATIC:
                    {
                        // Note: If the DIM parameters are not constants, we
                        //       should probably make the array automatically $DYNAMIC.

                        bool is_shared = false;
                        bool is_static = false;
                        if(tok != STATIC && tk(SHARED).parse(scan))
                        {
                            if(in_sub)
                                throw InvalidateAction("DIM SHARED invalid in SUB/FUNCTION");
                            // all vars in this definition will be shared
                            is_shared = true;
                        }
                        if(tok == STATIC)
                        {
                            if(!in_sub)
                                throw InvalidateAction("STATIC invalid outside SUB/FUNCTION");

                            is_static = true;
                        }
                        if(!is_shared)
                        {
                            /* If the SUB is declared STATIC, set is_static */
                            if(in_sub)
                            {
                            }
                            if(in_sub && in_sub->is_static)
                                is_static = true;
                        }
#ifdef BINDING_DEBUG
                        fprintf(stderr, "DIM: %s, %s, %s\n",
                            is_shared ? "Shared" : "Not shared",
                            is_static ? "Static" : "Not static",
                            in_sub ? (in_sub->is_static ? "Static sub" : "Non-static sub")
                                   : "Not sub");
#endif
                        /* DIM and REDIM.
                         * Note: REDIM can only be applied to $DYNAMIC arrays,
                         *             not scalars, utypes or $STATIC arrays.
                         *
                         * Note: If the array has not previously been declared,
                         *       declare it as $DYNAMIC if REDIM was used even if $STATIC was active.
                         */

                        do {
                            LambdaExpectToken(IDENTIFIER, return false);
                            std::string varname = scan.GetInfo().str;
                            std::vector<std::pair<exprp,exprp> > dim_exprs;

                            if(tk('(').parse(scan))
                            {
                                /* TODO: Should also support mere () for STATIC */
                                do {
                                    LambdaExpectNumExpr((expr), return false);
                                    exprp first  = new expr_number((long)option_base, TypeInt);
                                    exprp second = ctx.exp_result;
                                    if(tk(TO).parse(scan))
                                    {
                                        first = second;
                                        LambdaExpectNumExpr((expr), return false);
                                        second = ctx.exp_result;
                                    }

                                    dim_exprs.push_back(std::make_pair(first, second));
                                } while(tk(',').parse(scan));
                                LambdaExpectToken(')', return false);
                            }

                            TypeP type;

                            if(tk(AS).parse(scan))
                            {
                                LambdaExpectSubRule((this->type), return false);
                                type = ctx.typ_result;
#ifdef BINDING_DEBUG
                                dprintf(dp_declare, "DIM AS type: %s\n",
                                    type ? type->Dump().c_str() : "NULL");
#endif
                            }

                            if(!Do_Dim(lexloc, varname,
                                       dim_exprs,
                                       type,
                                       tok == REDIM,
                                       is_shared,
                                       is_static))
                            {
                                throw InvalidateAction("Invalid function call");
                            }
                        } while(tk(',').parse(scan));
                        return true;
                    }

                    case ERASE:
                    {
                        /* For static arrays, ERASE clears their contents to zero.
                         * For dynamic arrays, ERASE shrinks them to zero.
                         * For magic arrays, it delegates the work.
                         */
                        do {
                            if(!tq(IDENTIFIER).parse(scan)) break;
                            const std::string arrayname = scan.GetInfo().str;
                            VarContext c = FindVariableFromScope(arrayname, true);
                            if(!c.info)
                                throw InvalidateAction("Array not defined");

                            expr_var* v = new expr_var(arrayname);
                            v->set_type(c.info->type);
                            v->setwholearray();
                            SetStmt(new stmt_erase(lexloc, v));
                        } while(tk(',').parse(scan));
                        return true;
                    }

                    case OPEN:
                    {
                        LambdaExpectStringExpr((expr),  return false );
                        exprp param1 = ctx.exp_result; /* 1:filename, 2:mode */

                        QBtoken mode = RANDOM;
                        int access=0;  // 1=read, 2=write
                        int locking=0; // 0=shared, 1=read, 2=write
                        bool access_defined=false;
                        bool locking_defined=false;
                    OpenLoop:
                        switch(scan.get())
                        {
                            case FOR:
                                switch(mode=scan.get())
                                {
                                    case RANDOM: access=3; goto OpenLoop;
                                    case OUTPUT: access=2; goto OpenLoop;
                                    case INPUT:  access=1; goto OpenLoop;
                                    case APPEND: access=2; goto OpenLoop;
                                    case BINARY: access=3; goto OpenLoop;
                                    default:
                                        throw InvalidateAction("Expected RANDOM, OUTPUT, INPUT, APPEND or BINARY");
                                }
                            case ACCESS:
                            {
                                if(access_defined) throw InvalidateAction("Duplicate ACCESS clause");
                                int allow_access=0;
                                iterator_t
                                it = scan.pos(); if(scan.get() == READ)  allow_access|=1; else scan.revert(it);
                                it = scan.pos(); if(scan.get() == WRITE) allow_access|=2; else scan.revert(it);
                                if(!allow_access) throw InvalidateAction("Expected READ or WRITE");
                                access &= allow_access;
                                if(!access) throw InvalidateAction("File must be OPENed for a purpose");
                                access_defined = true;
                                goto OpenLoop;
                            }
                            case SHARED:
                                if(locking_defined) throw InvalidateAction("Duplicate LOCK or SHARED clause");
                                locking_defined=true;
                                locking=0; goto OpenLoop;
                            case LOCK:
                            {
                                iterator_t it = scan.pos();
                                if(locking_defined) throw InvalidateAction("Duplicate LOCK or SHARED clause");
                                if(scan.get() == READ) locking|=1; else scan.revert(it);
                                it = scan.pos();
                                if(scan.get() == WRITE) locking|=2; else scan.revert(it);
                                if(!locking) throw InvalidateAction("Expected READ or WRITE");
                                locking_defined=true;
                                goto OpenLoop;
                            }
                            case AS:  break;  // it's syntax 1
                            case ',': break; // it's syntax 2
                            default: return false;
                        }
                        bool oldsyntax = access == 0;

                        LambdaExpectSubRule((autofilenum), return false);
                        exprp filenum = ctx.exp_result;

                        exprp reclen;
                        exprp filename = param1;
                        if(oldsyntax)
                        {
                            if(param1->is_const())
                            {
                                /* Parse param1, get access mode (O/I/R/B/A) */
                                std::string m = (dynamic_cast<expr_string*>(&*param1))->get();
                                /**/ if(m=="O") { mode=OUTPUT; access=2; oldsyntax=false; }
                                else if(m=="I") { mode=INPUT;  access=1; oldsyntax=false; }
                                else if(m=="R") { mode=RANDOM; access=3; oldsyntax=false; }
                                else if(m=="B") { mode=BINARY; access=3; oldsyntax=false; }
                                else if(m=="A") { mode=APPEND; access=2; oldsyntax=false; }
                                else throw InvalidateAction("Expected string expression, or \"O\", \"I\", \"R\", \"B\" or \"A\"");
                            }

                            LambdaExpectToken(',', return false);

                            LambdaExpectStringExpr((expr),  return false );
                            filename = ctx.exp_result;

                            if(tk(',').parse(scan))
                            {
                                LambdaExpectNumExpr((expr),  return false );
                                reclen = ctx.exp_result;
                            }
                        }
                        else if(tk(LEN).parse(scan))
                        {
                            LambdaExpectToken('=', return false);
                            LambdaExpectNumExpr((expr),  return false );
                            reclen = ctx.exp_result;
                        }

                        // We now have:
                        //  mode (token)
                        //    or param1 (string expr), which should be parsed
                        //  access (bitmask)
                        //  locking (bitmask)
                        //  filename (string expr)
                        //  reclen (null, or int expr)

                        if(oldsyntax)
                            SetStmt(new stmt_fileopen(lexloc,
                                filename,filenum, param1, access, locking, reclen));
                        else
                            SetStmt(new stmt_fileopen(lexloc,
                                filename,filenum, mode,  access, locking, reclen));
                        return true;
                    }

                    case CLOSE:
                    {
                        bool found=false;
                        if((autofilenum).parse(scan))
                            for(;;)
                            {
                                SetStmt(new stmt_lib_stmt(lexloc, tok, ctx.exp_result, "N"));
                                found=true;
                                if(!tk(',').parse(scan)) break;
                                LambdaExpectSubRule((autofilenum), return false);
                            }
                        if(!found) /* empty list = close all files */
                            SetStmt(new stmt_lib_stmt(lexloc, tok));
                        return true;
                    }

                    case FIELD:
                    {
                        LambdaExpectSubRule((autofilenum), break);
                        LambdaExpectToken(',', break);
                        exprp filenum = ctx.exp_result;
                        SetStmt(new stmt_lib_stmt(lexloc, tok, filenum, "N"));
                        for(;;)
                        {
                            LambdaExpectNumExpr((expr), return false);
                            exprp width = ctx.exp_result;
                            LambdaExpectToken(AS, return false);
                            LambdaExpectSubRule((var_target), break);
                            exprp target = ctx.exp_result;
                            TypeCheckString(target);

                            SetStmt(new stmt_lib_stmt(lexloc, tok, filenum,width,target, "NNV"));

                            if(!tk(',').parse(scan)) break;
                        }
                        return true;
                    }
                    case LPRINT:
                    case PRINT:
                    case PRINT_F:
                    case WRITE:
                    {
                        autoptr<stmt_print> print_stmt = new stmt_print(lexloc);
                        if(tok == LPRINT) print_stmt->SetLPRINT();

                        if(tok == PRINT_F || tok == WRITE)
                        {
                            LambdaExpectSubRule((autofilenum), break);
                            LambdaExpectToken(',', break);

                            print_stmt->SetFile(ctx.exp_result);
                        }

                        exprp was_using;
                        if(tok != WRITE && tk(USING).parse(scan))
                        {
                            LambdaExpectStringExpr((expr), break);
                            LambdaExpectToken(';', break);
                            was_using = ctx.exp_result;

                            print_stmt->SetFormat(was_using);
                        }

                        bool need_delim = false;
                        bool ends_bare  = true;
                        for(bool first=true; ; first=false)
                        {
                            if(tk(',').parse(scan))
                            {
                                if(!was_using && tok != WRITE) // not in USING/WRITE
                                {
                                    print_stmt->AppendTab();
                                }
                                need_delim = false;
                                ends_bare  = false;
                                continue;
                            }
                            if(tk(';').parse(scan))
                            {
                                need_delim = false;
                                ends_bare  = false;
                                continue;
                            }
                            /* Note: need_delim not actually used here.
                               PRINT "a"b$"c" is allowed by GW-BASIC, and
                                this backward compatibility doesn't hurt.. much.
                             */
                            if(tok != WRITE)
                            {
                                if(tk(SPC).parse(scan))
                                {
                                    LambdaExpectToken('(', return false);
                                    LambdaExpectNumExpr((expr), return false);
                                    LambdaExpectToken(')', return false);

                                    print_stmt->AppendSPCfunc(ctx.exp_result);

                                    need_delim = true;
                                    ends_bare  = true;
                                    continue;
                                }
                                if(tk(TAB).parse(scan))
                                {
                                    LambdaExpectToken('(', return false);
                                    LambdaExpectNumExpr((expr), return false);
                                    LambdaExpectToken(')', return false);

                                    print_stmt->AppendTABfunc(ctx.exp_result);

                                    need_delim = true;
                                    ends_bare  = true;
                                    continue;
                                }
                            }
                            if(tok == WRITE && !first)
                            {
                                print_stmt->AppendCommaCh();
                            }
                            if(!(expr).parse(scan))
                            {
                                // USING requires at least one arg.
                                if(was_using && first) return false;
                                break;
                            }
                            exprp param = ctx.exp_result;

                            if(tok == WRITE && param->is_string())
                            {
                                print_stmt->AppendQuoteCh();
                                print_stmt->AppendExpr(param);
                                print_stmt->AppendQuoteCh();
                            }
                            else
                            {
                                if(tok != WRITE && !was_using && param->is_num())
                                {
                                    print_stmt->AppendSpace();
                                    print_stmt->AppendExpr(param);
                                    print_stmt->AppendSpace();
                                }
                                else
                                {
                                    expr_builtin_string* s =
                                        dynamic_cast<expr_builtin_string*> (&*param);
                                    if(s) /* Smart translations for several nn$ builtins */
                                    {
                                        if(s->GetToken() == STR_S && !was_using)
                                        {
                                            /* STR$(num) prints a space and num */
                                            print_stmt->AppendSpace();
                                            param = s->GetParam();
                                        }
                                        else if(s->GetToken() == LTRIM_S && !was_using)
                                        {
                                            expr_builtin_string* s2 =
                                                dynamic_cast<expr_builtin_string*> (&*s->GetParam());
                                            if(s2->GetToken() == STR_S)
                                            {
                                                /* LTRIM$(STR$(num)) prints just the num */
                                                param = s2->GetParam();
                                            }
                                        }
                                        else if(s->GetToken() == CHR_S)
                                        {
                                            /* CHR$(numconst) can be translated
                                             * to a string. */
                                            expr_number* n =
                                                dynamic_cast<expr_number*> (&*s->GetParam());
                                            if(n)
                                            {
                                                long long int i = n->get_i();
                                                if(i < 0 || i > 255)
                                                    throw InvalidateAction("Invalid parameter for CHR$");
                                                char Buf[1] = { (char) i };
                                                param = new expr_string( std::string(Buf,Buf+1) );
                                            }
                                        }
                                    }
                                    print_stmt->AppendExpr(param);
                                }
                            }

                            need_delim = true;
                            ends_bare  = true;
                        }
                        if(ends_bare || tok==WRITE)
                            print_stmt->AppendNewline();

                        SetStmt(&*print_stmt);
                        return true;
                    }

                    case INPUT:
                    case INPUT_F:
                    case LINEINPUT:
                    case LINEINPUT_F:
                    {
                        autoptr<stmt_input> input_stmt = new stmt_input(lexloc);

                        const bool is_file = tok == INPUT_F   || tok == LINEINPUT_F;
                        const bool is_line = tok == LINEINPUT || tok == LINEINPUT_F;

                        if(is_line) input_stmt->SetLINEINPUT();
                        if(is_file)
                        {
                            LambdaExpectSubRule((autofilenum), break);
                            LambdaExpectToken(',', break);
                            input_stmt->SetFile(ctx.exp_result);
                        }
                        else
                        {
                            if(tk(';').parse(scan))
                            {
                                input_stmt->SetKeepCursor();
                            }
                            if(tk(STRINGCONST).parse(scan))
                            {
                                DoString(scan);
                                input_stmt->SetPrompt(ctx.exp_result);
                                if(!(tk(';') | tk(',')).expected().parse(scan))
                                    break;
                            }
                        }
                        if(is_line)
                        {
                            LambdaExpectStringExpr((var_target), break);
                            input_stmt->AppendTarget(ctx.exp_result);
                        }
                        else
                        {
                            do {
                                LambdaExpectSubRule((var_target), return false);
                                input_stmt->AppendTarget(ctx.exp_result);
                            } while(tk(',').parse(scan));
                        }
                        SetStmt(&*input_stmt);
                        return true;
                    }

                    case ON:
                    {
                        LambdaExpectNumExpr((expr), break);

                        autoptr<stmt_select> select_stmt = new stmt_select(lexloc);
                        select_stmt->ref =
                            new expr_binary((QBtoken)'-', ctx.exp_result, new expr_number(1));

                        bool is_goto = false;
                        if(tk(GOSUB).parse(scan)) is_goto=false;
                        else if(tk(GOTO).parse(scan)) is_goto=true;
                        else throw InvalidateAction("Expected: GOTO or GOSUB");

                        for(int value=0; ; ++value)
                        {
                            lexloc = scan.GetLoc();

                            LambdaExpectSubRule((anylabeltarget), break);
                            exprp tg = ctx.exp_result;
                            exprp ex = new expr_caseis(
                                (QBtoken)'=', new expr_number(value));

                            if(is_goto)
                                select_stmt->cases.push_back(
                                    std::make_pair(ex, new stmt_labelref(lexloc, GOTO, tg)));
                            else
                                select_stmt->cases.push_back(
                                    std::make_pair(ex, new stmt_labelref(lexloc, GOSUB, tg)));

                            if(!tk(',').parse(scan)) break;
                        }
                        SetStmt(&*select_stmt);
                        return true;
                    }
                }
                return false;
            }
        )

    |   /* Assign statements without "LET" */
        LambdaParser([&] (scanner_t& scan) -> bool
        {
            if(!(var_target).parse(scan)) return false;
            LexLocation lexloc = scan.GetLoc();
            exprp target = ctx.exp_result;
            if(!tk('=').parse(scan)) return false;
            LambdaExpectSubRule((expr), return false);
            exprp source = ctx.exp_result;

            /* TODO: Check type more carefully */
            TypeP tt = target->GetType(), st = source->GetType();
            if(tt && st
            && ((tt->is_array()
              != st->is_array())
             || (tt->is_string()
              && st->is_num())
             || (tt->is_num()
              && st->is_string())))
            {
                throw InvalidateAction("Type mismatch");
            }
            SetStmt(new stmt_let(lexloc, target, source));
            return true;
        })
    |   /* SUB call without "CALL" */
        LambdaParser([&] (scanner_t& scan) -> bool
        {
            LexLocation lexloc = scan.GetLoc();
            if(!tk(IDENTIFIER).parse(scan)) return false;

            const std::string subname = scan.GetInfo().str;
            // TODO: check namespace

            std::vector<exprp> params;
            if((func_param).parse(scan))
                for(;;)
                {
                    params.push_back(ctx.exp_result);
                    if(!tk(',').parse(scan)) break;
                    LambdaExpectSubRule((func_param), return false);
                }
            SetStmt(new stmt_subcall(lexloc, subname, params));
            return true;
        });

    func_param = LambdaParser(
        [&] (scanner_t& scan) -> bool
        {
            iterator_t it = scan.pos();
            if(tk(IDENTIFIER).parse(scan))
            {
                const std::string arrayname = scan.GetInfo().str;
                if(tk('(').parse(scan)
                && tk(')').parse(scan))
                {
                    // Magic array!
                    VarContext c = FindVariableFromScope(arrayname, true);
                    if(!c.info)
                        throw InvalidateAction("Array not defined");
                    expr_var* v = new expr_var(arrayname);
                    v->set_type(c.info->type);
                    v->setwholearray();
                    SetExpr(v);
                    return true;
                }
            }
            scan.revert(it);
            return (expr).parse(scan);
        }, "SUB/FUNCTION parameter");


    expr = LambdaParser(
        [&] (scanner_t& scan) -> bool
        {
            iterator_t it = scan.pos();

            /* PRIORITY LEVELS:
             *
             *  1 IMP
             *  2 EQV
             *  3 XOR
             *  4 OR
             *  5 AND
             *  6 unary NOT
             *  7 comparisons
             *  8 + -
             *  9 MOD  _SHL _SHR
             * 10 \
             * 11 * /
             * 12 unary -  (using token "IF" here)
             * 13 ^
             */
            unsigned prio = 0;
            std::vector<ExprPrioItem> priostack;

            #define BINARY_OP(level, op) \
                while(prio >= level) PRIO_RETURN(); \
                PRIO_ENTER_OP(level, op)

            auto PRIO_ENTER_OP = [&prio,&priostack] (int level, QBtoken op)
            {
                /*
                fprintf(stderr, "prio=%u, enter %u(%s)\n",
                    (prio), level,
                    GetTokenName(op).c_str());
                */
                ExprPrioItem tmp;
                tmp.prio = (prio);
                tmp.tok = op;
                tmp.exp = ctx.exp_result;
                prio = level;
                priostack.push_back(tmp);
            };

            auto PRIO_RETURN = [&prio,&priostack] ()
            {
                if(priostack.empty())
                {
                    fprintf(stderr, "error: PRIO_RETURN called when empty (prio=%u)\n",
                        (prio));
                }
                const ExprPrioItem& top = priostack.back();
                switch(top.tok)
                {
                    case NOT:
                        //fprintf(stderr, "Creating NOT <...>\n");
                        if(!ctx.exp_result->is_num()) goto TypeMismatch;
                        SetExpr(new expr_unary(NOT, ctx.exp_result));
                        break;
                    case IF: // unary minus
                        //fprintf(stderr, "Creating MINUS <...>\n");
                        if(!ctx.exp_result->is_num()) goto TypeMismatch;
                        SetExpr(new expr_unary((QBtoken)'-', ctx.exp_result));
                        break;
                    default:
#ifdef PARSER_DEBUG
                        fprintf(stderr, "Creating %s <%s>, <%s>\n",
                            GetTokenName(top.tok).c_str(),
                            typenameof(*top.exp).c_str(),
                            typenameof(*ctx.exp_result).c_str()
                        );
#endif
                        bool string_ok = false;
                        switch(top.tok)
                        {
                            case (QBtoken)'+': case (QBtoken)'<': case (QBtoken)'>': case (QBtoken)'=':
                            case NOTEQUAL: case LESSOREQUAL: case GREATEROREQUAL:
                                string_ok = true;
                            default: break;
                        }
                        if(string_ok)
                        {
                            if(ctx.exp_result->is_num()
                            != top.exp->is_num()) goto TypeMismatch;
                        }
                        else
                        {
                            if(top.exp->is_num().is_false()) goto TypeMismatch;
                            if(ctx.exp_result->is_num().is_false()) goto TypeMismatch;
                        }
                        switch(top.tok)
                        {
                            case IMP: // a IMP b = b OR NOT a
                              { exprp p = ctx.exp_result;
                                SetExpr(new expr_unary(NOT, top.exp));
                                SetExpr(new expr_binary(OR, ctx.exp_result, p));
                                break; }
                            case EQV: // a EQV b = NOT(a XOR b)
                              { SetExpr(new expr_binary(XOR, top.exp, ctx.exp_result));
                                SetExpr(new expr_unary(NOT, ctx.exp_result));
                                break; }
                            default:
                                SetExpr(new expr_binary(top.tok, top.exp, ctx.exp_result));
                        }
                        break;
                }
                prio = top.prio;
                priostack.pop_back();
                return;
            TypeMismatch:
                throw InvalidateAction("Type mismatch\n");
            };

            bool first = true;
        Continues:
            if(!scan.eof()) /* check unary */
            {
                iterator_t unary_it = scan.pos();
                switch( (int) scan.get())
                {
                    case NOT: PRIO_ENTER_OP(6, NOT); goto Continues;
                    case '-': PRIO_ENTER_OP(6, IF); goto Continues;
                    default: scan.revert(unary_it);
                }
            }
            if(first)
                { first=false; if(!(expr_meat).parse(scan)) goto Fail; }
            else
                { LambdaExpectSubRule((expr_meat), goto Fail); }
        Continued:
            if(!scan.eof()) /* check binary */
            {
                iterator_t opr_it = scan.pos();
                QBtoken opr = scan.get();
                switch( (int) opr)
                {
                    case IMP: BINARY_OP(1, opr); goto Continues;
                    case EQV: BINARY_OP(2, opr); goto Continues;
                    case XOR: BINARY_OP(3, opr); goto Continues;
                    case OR:  BINARY_OP(4, opr); goto Continues;
                    case AND: BINARY_OP(5, opr); goto Continues;
                    // unary NOT
                    case NOTEQUAL: case LESSOREQUAL: case GREATEROREQUAL:
                    case '>': case '<': case '=':
                              BINARY_OP(7, opr); goto Continues;
                    case '+': case '-':
                              BINARY_OP(8, opr); goto Continues;
                    case MOD: BINARY_OP(9, opr); goto Continues;
                    case qbcSHL: BINARY_OP(9, opr); goto Continues;
                    case qbcSHR: BINARY_OP(9, opr); goto Continues;
                    case '\\': BINARY_OP(10, opr); goto Continues;
                    case '*': case '/':
                              BINARY_OP(11, opr); goto Continues;
                    // unary -
                    case '^': BINARY_OP(13, opr); goto Continues;

                    case NUMBER:
                        // Handle x = x-3 where -3 is a NUMBER token
                        if(scan.GetInfo().had_sign)
                        {
                            exprp tmp = ctx.exp_result;
                            DoNumber(scan);
                            exprp e = ctx.exp_result;
                            ctx.exp_result = tmp;
                            BINARY_OP(8, (QBtoken)'+');
                            SetExpr(e);
                            goto Continued;
                        }
                        break;

                    default: break;
                }
                scan.revert(opr_it);
            }
            while(!priostack.empty()) { PRIO_RETURN(); }
            return true;
        Fail:
            scan.revert(it);
            return false;
        }, "expression");

    expr_meat =
        '(' >> expr >> ')'

    |   LambdaTokParser(
        [&] (QBtoken tok, scanner_t& scan) -> bool
        {
            const char* pattern = 0;
            /* LIBFUNCTIONS */
            switch(tok)
            {
                default:break;
                got_expr_builtin_numeric:
                    return (ParseLibFunction<expr_builtin_numeric>)((*this), scan, tok, pattern);
                got_expr_builtin_string:
                    return (ParseLibFunction<expr_builtin_string>)((*this), scan, tok, pattern);

                #define tokcase(s,p) case s: Do(p)
                #define Do(p) { pattern=p; goto got_expr_builtin_numeric; }

                     // numeric -> numeric functions
#include "libdata-expr_numeric.inc"

                #undef Do
                #define Do(p) { pattern=p; goto got_expr_builtin_string; }

#include "libdata-expr_string.inc"

                #undef Do
                #undef tokcase

                case LBOUND:
                case UBOUND:
                {
                    LambdaExpectToken('(', break);
                    LambdaExpectToken(IDENTIFIER, break);
                    const std::string arrayname = scan.GetInfo().str;
                    VarContext c = FindVariableFromScope(arrayname, true);
                    if(!c.info)
                        throw InvalidateAction("Array not defined");

                    expr_var* v = new expr_var(arrayname);
                    v->set_type(c.info->type);
                    v->setwholearray();
                    exprp p1 = v;
                    ctx.exp_result = 0;
                    if(tk(',').parse(scan))
                    {
                        LambdaExpectNumExpr((expr), break);
                    }
                    LambdaExpectToken(')', break);
                    SetExpr(new expr_lubound(tok, p1, ctx.exp_result));
                    return true;
                }

                case NUMBER:
                    { DoNumber(scan); return true; }
                case STRINGCONST:
                    { DoString(scan); return true; }
            }
            return false;
        }, "a constant or a built-in function")

    |   /**** RVAL */
        /* Var or const or function call.
         * Collect it into an expr_var first. Then decide if to convert
         * it to expr_userfun.
         *
         * Function calls look just like array variables.
         * Only context will tell whether it should be converted into a FUNCTION.
         */
        LambdaParser([this] (scanner_t& scan) -> bool
        {
            return (this->ParseVarDef)(scan,
                [&] (expr_var*, exprp c) -> bool
                { // const
                    SetExpr(c);
                    return true;
                },

                [&] (expr_var* var, SubFuncP s, bool) -> bool
                { // sub
                    if(!s->returntype)
                    {
                        throw InvalidateAction("Invalid use of a SUB");
                    }
                    SetExpr(&*var->convert_to_functioncall(s->returntype));
                    return true;
                }
            );
        }
        , "rvalue");

    ambiguous_id = LambdaParser([this] (scanner_t& scan) -> bool
    {
        iterator_t it = scan.pos();

        /* This merely parses stuff, without context or type info */

        autoptr<expr_var> root  = 0;
        autoptr<expr_var> pivot = 0;

        bool is_fn = false;

        do {
            if(tk(FNIDENTIFIER).parse(scan))
            {
                if(root)
                    throw InvalidateAction("TYPEs cannot contain FN members");
                is_fn = true;
            }
            else
            {
                if(!tk(IDENTIFIER).parse(scan)) goto Fail;
                if(is_fn)
                    throw InvalidateAction("FNs cannot contain TYPE elements");
            }

            std::string identifier = scan.GetInfo().str;

            if(root)
            {
                if(HasTypeSuffix(identifier))
                    throw InvalidateAction("TYPE member cannot have a type suffix");
            }

            autoptr<expr_var> v = new expr_var(identifier);

            if(tk('(').parse(scan))
            {
                /* indexes */
                do {
                    LambdaExpectAnyExpr((func_param), goto Fail);
                    v->addindex(ctx.exp_result);
                } while(tk(',').parse(scan));

                LambdaExpectToken(')', goto Fail);
            }

            if(pivot)
            {
                /* It's a member. Therefore this whole construct
                 * is an user type. User types cannot contain type suffixes.
                 */
                pivot->addmember(v);
                pivot = v;
            }
            else
            {
                root  = v;
                pivot = v;
            }

        } while(tk('.').parse(scan));

        SetExpr(&*root);
        return true;
    Fail:
        scan.revert(it);
        return false;
    });

program.setname("program");
separator.setname("separator");
type.setname("type");
anylabeltarget.setname("anylabeltarget");
plain_coordinate.setname("plain_coordinate");
step_coordinate.setname("step_coordinate");
longstatement.setname("longstatement");
shortstatement.setname("shortstatement");
shortif.setname("shortif");
casematch.setname("casematch");
simplestatement.setname("simplestatement");
expr_meat.setname("expr_meat");
var_target.setname("var_target");
autofilenum.setname("autofilenum");
ambiguous_id.setname("ambiguous_id");
expr.setname("expr");
func_param.setname("func_param");
    }

    bool parse(scanner_t& scanner) const
    {
        return program.parse(scanner);
    }

    #undef expr
};

// lowercase means optional
// S       string
// N       numeric
// X       string or numeric
// L       anylabeltarget
// R       anylabeltarget where zero constitutes as no parameter
// P       step coordinate                                              param: flag,x,y
// C       coordinate (no step)                                         param: x,y
// V       var_target
// VS      var_target that must be a string
// VN      var_target that must be numeric
// A       array with optional indexes
// F       file number
// ZG      GOSUB                                     (ON PLAY/KEY/TIMER)
// ZH      LIST/OFF/ON                               (KEY)              param: 0/-1/1
// ZK      STOP/OFF/ON                               (PLAY,TIMER)       param: 0/-1/1
// ZL      B/BF                                      (LINE)             param: 1/2 [0]
// ZO      GOTO                                      (ON ERROR)
// ZP      XOR/PSET/PRESET/AND/OR, XOR being default (PSET/PRESET)      param: 0/1/2/3/4 [0]
// ZS      SCREEN                                    (VIEW)
// ZT      TO                                        (VIEW PRINT)

// ,-=()   literals
// %       any >=1 number of following separated by comma
//
// |       another alternative for entire syntax
// ?       if next token is not matched, parsing skips successfully
//         to the next closing parenthesis or end of pattern
//
//
// spaces are ignored
template<typename RetType>
static bool ParseLibItem(QB_parser& parser, scanner_t& scan, QBtoken tok,
                         const char* defs, RetType*& result,
                         std::string& matched_pattern)
{
    std::vector<exprp> params;
    std::vector<bool> lvaluemap;
    iterator_t it = scan.pos();

    LexLocation lexloc = scan.GetLoc();

    #define MATCH()   matched_pattern += std::toupper(MatchType)
    #define NOMATCH() matched_pattern += std::tolower(MatchType)
    /*
      Matched pattern keywords:
        S = simple (string)
        N = simple (num)
        X = S or N
        L = label
        P = stepcoordinate (three nums)
        C = coordinate (two nums)
        V = lvalue
    */

    const char* loopbegin = 0; bool loop_opt = false;
    matched_pattern.clear();
  reloop:
    while(*defs)
    {
        bool optional = loop_opt;
        char MatchType = 0;
    reswitch:
        iterator_t curit = scan.pos();
        if(*defs == '|') break;
        //fprintf(stderr, "'%c' (%s) (%s)?\n", *defs, defs, GetTokenName(tok).c_str());
        switch(*defs)
        {
            case ' ': ++defs; goto reswitch;
            case '?':
                optional = true; ++defs; goto reswitch;
            case 'S': case 's':
                MatchType='S';
                if(!parser.expr.parse(scan)
                || ctx.exp_result->is_string().is_false())
                { if(*defs=='S')goto Failed;
                IgnoreFail: scan.revert(curit); NOMATCH();
                  params.push_back(0); break; }
                MATCH(); params.push_back(ctx.exp_result); break;
            case 'N': case 'n':
                MatchType='N';
                if(!parser.expr.parse(scan)
                || ctx.exp_result->is_num().is_false())
                { if(*defs=='N')goto Failed; goto IgnoreFail; }
                MATCH(); params.push_back(ctx.exp_result); break;
            case 'X': case 'x':
                MatchType='X';
                if(!parser.expr.parse(scan))
                { if(*defs=='X')goto Failed; goto IgnoreFail; }
                MATCH(); params.push_back(ctx.exp_result); break;
            case 'L': case 'l':
                MatchType='L';
                if(!parser.anylabeltarget.parse(scan))
                { if(*defs=='L')goto Failed; goto IgnoreFail; }
                MATCH(); params.push_back(ctx.exp_result); break;
            case 'R': case 'r':
                MatchType='L';
              { if(!parser.anylabeltarget.parse(scan))
                { if(*defs=='R')goto Failed; goto IgnoreFail; }
                expr_label* labelp = dynamic_cast<expr_label*>(&*ctx.exp_result);
                if(labelp && labelp->get_label().empty()) ctx.exp_result=0;
                if(ctx.exp_result) { MATCH(); } else { NOMATCH(); }
                params.push_back(ctx.exp_result); break; }
            case 'P': case 'p':
                MatchType='P';
              { if(!parser.step_coordinate.parse(scan))
                { if(*defs=='P')goto Failed; goto IgnoreFail; }
                MATCH();
                params.push_back(new expr_number(coord_res.step));
                params.push_back(coord_res.x);
                params.push_back(coord_res.y);
                break; }
            case 'C': case 'c':
                MatchType='C';
              { if(!parser.plain_coordinate.parse(scan))
                { if(*defs=='C')goto Failed; goto IgnoreFail; }
                MATCH();
                params.push_back(coord_res.x);
                params.push_back(coord_res.y);
                break; }
            case 'V': case 'v':
                MatchType='V';
                if(!parser.var_target.parse(scan))
                { if(*defs=='V')goto Failed; goto IgnoreFail; }
                MATCH();
                params.push_back(ctx.exp_result);
                lvaluemap.resize(params.size());
                lvaluemap[params.size()-1] = true;
                if(defs[1]=='S') { ++defs; TypeCheckString(ctx.exp_result); }
                if(defs[1]=='N') { ++defs; TypeCheckNum(ctx.exp_result);    }
                break;
            case 'F': case 'f':
                MatchType='N';
                if(!parser.autofilenum.parse(scan))
                { if(*defs=='F')goto Failed; goto IgnoreFail; }
                MATCH();
                params.push_back(ctx.exp_result); break;
            case 'A': case 'a':
            {
                /* TODO */
                /* array name, possibly with starting offset */
                MatchType='A';
                if(!parser.expr.parse(scan))
                { if(*defs=='A')goto Failed; goto IgnoreFail; }
                MATCH(); params.push_back(ctx.exp_result); break;
                break;
            }
            case ',': case '-':
            case '=': case '(':
            case ')':
                if(!tk(*defs).parse(scan)) goto Failed;
                break;
            case '%':
                loopbegin = ++defs;
                loop_opt = optional;
                goto reswitch;
            case 'Z': case 'z':
            {
                switch(defs[1])
                {
                    case 'P': case 'p': // XOR/PSET/PRESET/AND/OR
                        MatchType='N';
                        MATCH();
                        if(tk(XOR).parse(scan)) params.push_back(new expr_number(0));
                        else if(tk(PSET).parse(scan)) params.push_back(new expr_number(1));
                        else if(tk(PRESET).parse(scan)) params.push_back(new expr_number(2));
                        else if(tk(AND).parse(scan)) params.push_back(new expr_number(3));
                        else if(tk(OR).parse(scan)) params.push_back(new expr_number(4));
                        else params.push_back(new expr_number(0));
                        ++defs; break;
                    case 'T': case 't': // TO
                        if(!tk(TO).parse(scan))
                            { Zfail: if(*defs=='Z') goto Failed; ++defs; goto IgnoreFail; }
                        ++defs; break;
                    case 'L': case 'l': // B/BF
                        MatchType='N';
                      { if(!tk(IDENTIFIER).parse(scan))
                            goto Zfail;
                        const std::string id = BasUCASE(scan.GetInfo().str);
                        if(id == "B") { MATCH(); params.push_back(new expr_number(1)); }
                        else if(id == "BF") { MATCH(); params.push_back(new expr_number(2)); }
                        else goto Zfail;
                        ++defs; break; }
                    case 'S': case 's': // SCREEN
                        MatchType='N';
                        if(tk(SCREEN).parse(scan)) params.push_back(new expr_number(1));
                        else params.push_back(new expr_number(0));
                        MATCH();
                        ++defs; break;
                    case 'K': case 'k': // STOP/OFF/ON
                        MatchType='N';
                        if(tk(STOP).parse(scan)) params.push_back(new expr_number(0));
                        else if(tk(OFF).parse(scan)) params.push_back(new expr_number(-1));
                        else if(tk(ON).parse(scan)) params.push_back(new expr_number(1));
                        else goto Zfail;
                        MATCH();
                        ++defs; break;
                    case 'H': case 'h': // LIST/OFF/ON
                        MatchType='N';
                        if(tk(LIST).parse(scan)) params.push_back(new expr_number(0));
                        else if(tk(OFF).parse(scan)) params.push_back(new expr_number(-1));
                        else if(tk(ON).parse(scan)) params.push_back(new expr_number(1));
                        else goto Zfail;
                        MATCH();
                        ++defs; break;
                    case 'G': case 'g': // GOSUB
                        if(!tk(GOSUB).parse(scan))
                            goto Zfail;
                        ++defs; break;
                    case 'O': case 'o': // GOTO
                        if(!tk(GOTO).parse(scan))
                            goto Zfail;
                        ++defs; break;
                }
            }
        }
        //fprintf(stderr, "- ok\n");
        if(false)
        {
        Failed:
            //fprintf(stderr, "- fail\n");
            if(optional)
            {
                while(*defs && *defs != '|' && *defs != ')') ++defs;
                if(*defs == ')') { continue; }
                if(!*defs) break;
            }
            while(*defs && *defs!='|') ++defs;
            if(*defs=='|')
            {
                ++defs;
                scan.revert(it);
                params.clear();
                lvaluemap.clear();
                matched_pattern.clear();
                continue;
            }
            return false;
        }
        ++defs;
    }
    if(loopbegin && tk(',').parse(scan))
    {
        defs = loopbegin;
        goto reloop;
    }

    result = new RetType(lexloc, tok);
    result->set_params(params);
    result->set_lvalue_map(lvaluemap);

    return true;
}

/************************************/

int main(int argc, char** argv)
{
    /*setbuf(stdin,NULL);*/
    const char* fn = 0;
    FILE* fp = stdin;
    if(argv[1])
    {
        fp = fopen(fn = argv[1], "rt");
    }

    yyStartFile(fp, fn ? fn : "<stdin>");

    Lex<QBtoken> lexer;
    scanner_t scan(lexer.begin(), lexer.end());
    QB_parser pa;

    pa.parse(scan);

    if(!scan.eof())
        yyerror("Incomplete parse\n", scan.pos().GetLoc());

    Compile(ctx.stm_result,
            global_data);

    return 0;
}
