#include "typenameof.hh"

#include "autoptr"
//#include <boost/shared_ptr.hpp>

template<typename ScannerT>
/* abstract parser */
struct rule_a : public ptrable
{
    rule_a() {}
    virtual ~rule_a() {}
    virtual bool parse(ScannerT& scan) const = 0;
    virtual const std::string what() const = 0;
    std::string name;
};
template<typename ScannerT, typename ParserT>
/* concrete parser */
class rule_c: public rule_a<ScannerT>, private ParserT
{
public:
    rule_c(const parser<ParserT>& parser)
        : rule_a<ScannerT>(), ParserT(parser.derived()) { }

    virtual bool parse(ScannerT& scan) const { return ParserT::parse(scan); }
    virtual const std::string what() const   { return ParserT::what(); }
};

#ifdef PARSER_DEBUG
#include <list>
#endif

#ifdef PARSER_DEBUG
template<typename ScannerT>
static void DumpStack(const std::string& newname, bool pop=false)
{
    static std::list<std::string> stack;
    if(pop) stack.pop_back();
    else stack.push_back(newname);

    const unsigned maxlen = 130;
    unsigned len = 0;
    for(std::list<std::string>::const_iterator
        i = stack.begin(); i != stack.end(); ++i)
    {
        if(len + i->size() >= maxlen)
        {
            dprintf(dp_rulecomego, "\n"); len = 0;
        }
        dprintf(dp_rulecomego, "> %s ", i->c_str());
        len += i->size() + 4;
    }
    dprintf(dp_rulecomego, "\n");
}
#endif

template<typename ScannerT, typename EnterContext=NopActor, typename ExitContext=NopActor>
class rule: public parser<rule<ScannerT,EnterContext,ExitContext> >
{
    typedef rule_a<ScannerT> abs_t;

    /* pointer 1, so that this object can be copied before assigning,
     * and it'll work (pointers will share the same target)
     *   i.e. rule a, b;
     *        b = ... >> a;  (b gets a copy of a)
     *        a = ...;       (a is assigned an expression)
     */
    /* pointer 2, so that virtual may do its work
    */

    typedef ptrable_autoptr<abs_t> ptr2_t;
    typedef autoptr<ptr2_t> ptr1_t;
    //typedef boost::shared_ptr<abs_t> ptr2_t;
    //typedef boost::shared_ptr<ptr2_t> ptr1_t;

    ptr1_t pars;

public:
    rule() : pars(new ptr2_t())
    {
    }
    rule(const rule& b) : pars(b.pars)
    {
    }
    template<typename A>
    rule(const parser<A>& parser) : pars( new ptr2_t() )
    {
        (*pars).reset( new rule_c<ScannerT, A> (parser) );
    }

    template<typename A>
    rule<ScannerT,EnterContext,ExitContext>&
    operator=(const parser<A>& parser)
    {
        (*pars).reset( new rule_c<ScannerT, A> (parser) );
        return *this;
    }

    void setname(const std::string& name) { (*pars)->name = name; }

    bool parse(ScannerT& scan) const
    {
        EnterContext() ();
#ifdef PARSER_DEBUG
        if(pars == 0) fprintf(stderr, "1ST LEVEL NULL??\n");
        if(*pars == 0) fprintf(stderr, "2ND LEVEL NULL??\n");
        const std::string& name = (*pars)->name;
        dprintf(dp_rulecomego, "Entering rule: %s (%s); stack now:\n", name.c_str(),
            typenameof(**pars).c_str());
        DumpStack<ScannerT> (name);
#endif
        bool ret = (*pars)->parse(scan);
#ifdef PARSER_DEBUG
        dprintf(dp_rulecomego, "Exiting rule: %s (%s); stack now:\n", name.c_str(),
            ret ? "success" : "no match");
        DumpStack<ScannerT> ("",true);
#endif
        ExitContext() ();
        return ret;
    }

    const std::string what() const
    {
        const std::string& nm = (*pars)->name;
        return !nm.empty() ? nm : (*pars)->what();
    }
};
