#include <list>
#include <cmath>

#include "play.hh"
#include "error.hh"
#include "exec.hh"
#include "stream.hh"

#include "Blip_Buffer.h"

void* PlayHandler = 0;

namespace
{
    bool Background = false;

    enum { SynthRate = 44100 };

    class Player
    {
    private:
        struct Noterec
        {
            int    note;
            int    volume;
            double duration;
        };
        std::list<Noterec> Buffer;

        Blip_Buffer                       Blip;
        Blip_Synth<blip_high_quality,100> Synth;
        std::vector<blip_sample_t> Audio;
        double time;
    public:
        Player() : time(0.0)
        {
            // The longest blip we can produce is at T32 L1 with infinite number of dots.
            // This makes  240/(32*1)*2 seconds  = 15 seconds.
            Blip.set_sample_rate(SynthRate, 15000);
            Synth.output(&Blip);
        }

        // Duration = length of note in seconds
        void QueueNote(int notenum, int volume, double duration)
        {
            Noterec tmp;
            tmp.note     = notenum;
            tmp.volume   = volume;
            tmp.duration = duration;
            Buffer.push_back(tmp);
        }
        void Flush()
        {
            // frequency = (2^(note/12))*440
            // wavelen   = SynthRate / ((2^(note/12))*440)
            int amplitude = -1;
            while(!Buffer.empty())
            {
                Noterec note = Buffer.front();
                Buffer.pop_front();

                Synth.volume( note.volume );

                double end_time = time + note.duration * SynthRate;

                double freq_hz = 440.0 * std::pow(2, (note.note-33) / 12.0);
                double wave_len;
                if(note.volume && note.note)
                {
                    wave_len = SynthRate / freq_hz;
                }
                else
                {
                    wave_len = end_time - time;
                    amplitude = 0;
                }

                while(time < end_time)
                {
                    Synth.update( (blip_time_t) time , amplitude );
                    time      += wave_len;
                    amplitude = -amplitude;
                }
                time = end_time;

                long count    = Blip.samples_avail();
                size_t endpos = Audio.size();
                Audio.resize(endpos + count);
                Blip.read_samples(&Audio[endpos], count);

                if(amplitude == 0)
                    amplitude = -1;
            }
        }
    };
    class PlayParser: public Player
    {
        std::string ParseBuffer;
        size_t      ParsePosition;
        int         Octave;
        int         Tempo;
        int         Length;
        int         Volume;
        int         Rhythm; // n/8
        int         Notenum, Notelength;
    public:
        PlayParser() : ParseBuffer(), ParsePosition(0),
                       Octave(2), Tempo(120), Length(4), Volume(50),
                       Rhythm(7),
                       Notenum(), Notelength()
        {
        }
        void Feed(const BAS_STR& pattern)
        {
            ParseBuffer += pattern;
            Parse();
        }
        void Parse()
        {
/*
  Commands:
    "O" <number>     Sets current octave (0-6)
    "L" <number>     Sets length, L1=whole,  L2=half, ...L64
    "T" <number>     Sets number of quarter notes per minute (32-255, 120 default)
    "V" <number>     Sets volume (0-15)
    "<"              Decrease octave by one
    ">"              Increase octave by one
    <note>
     [#+-]           Explicitly increase or decrease note by one
     [<number>]      Explicit length for that note
     [.]*            Explicitly add half of length (c = 1/1, c.. = 1/1 + 1/2 + 1/4)
    "MB"             Play on background
    "MF"             Play on foreground
    "MS"             Staccato (6/8 note + 2/8 pause)
    "MN"             Normal   (7/8 note + 1/8 pause)
    "ML"             Legato   (8/8 note + 0/8 pause)
    "X" <offset>     Include substring from that address (offset is that returned by VARPTR$(stringvariable))
                      Use VARPTR_ReadStr() from exec.hh
  Numbers:
    [0-9]+           Integer
    "=" <offset>     Include numeric value from that address (offset is that returned by VARPTR$(numvariable))
                      Use VARPTR_ReadNum() from exec.hh
  Notes:
    "N" <number>     Note n (0-84, 0 = rest)
    [ABCDEFG]        Note 0..6 + octave*12
    P                Note 0
*/
            while(ParsePosition < ParseBuffer.size())
            {
                switch(ParseBuffer[ParsePosition])
                {
                    case ' ':
                    case '\t':
                    case '\r':
                    case '\n':
                    case '\v':
                        ++ParsePosition;
                        break;
                    case 'O': case 'o':
                        ++ParsePosition;
                        Octave = (int) ParseNumber();
                        if(Octave < 0 || Octave > 6) B_ERROR(5);
                        break;
                    case 'L': case 'l':
                        ++ParsePosition;
                        Length = (int) ParseNumber();
                        if(Length < 1 || Length > 64) B_ERROR(5);
                        break;
                    case 'T': case 't':
                        ++ParsePosition;
                        Tempo = (int) ParseNumber();
                        if(Tempo < 32 || Tempo > 255) B_ERROR(5);
                        break;
                    case 'V': case 'v':
                        ++ParsePosition;
                        Volume = (int) ParseNumber();
                        if(Volume < 0 || Volume > 100) B_ERROR(5);
                        break;
                    case '<':
                        if(Octave > 0) --Octave;
                        ++ParsePosition;
                        break;
                    case '>':
                        if(Octave < 6) ++Octave;
                        ++ParsePosition;
                        break;
                    case 'M': case 'm':
                        ++ParsePosition;
                        if(ParsePosition >= ParseBuffer.size()) B_ERROR(5);
                        switch(ParseBuffer[ParsePosition++])
                        {
                            case 'B': case 'b': Background = true; break;
                            case 'F': case 'f': Background = false; break;
                            case 'S': case 's': Rhythm = 6; break;
                            case 'N': case 'n': Rhythm = 7; break;
                            case 'L': case 'l': Rhythm = 8; break;
                            default: B_ERROR(5);
                        }
                        break;
                    case 'N': case 'n':
                        ++ParsePosition;
                        Notenum = (int) ParseNumber();
                        Notelength = Length;
                        goto GotNote;
                    case 'P': case 'p':
                        ++ParsePosition;
                        Notenum = 0;
                        goto GetOptionalLength;
                    case 'C': case 'c': ++ParsePosition; Notenum = 1 + 0 + Octave*12; goto GetSharpFlat;
                    case 'D': case 'd': ++ParsePosition; Notenum = 1 + 2 + Octave*12; goto GetSharpFlat;
                    case 'E': case 'e': ++ParsePosition; Notenum = 1 + 4 + Octave*12; goto GetSharpFlat;
                    case 'F': case 'f': ++ParsePosition; Notenum = 1 + 5 + Octave*12; goto GetSharpFlat;
                    case 'G': case 'g': ++ParsePosition; Notenum = 1 + 7 + Octave*12; goto GetSharpFlat;
                    case 'A': case 'a': ++ParsePosition; Notenum = 1 + 9 + Octave*12; goto GetSharpFlat;
                    case 'B': case 'b': ++ParsePosition; Notenum = 1 + 11+ Octave*12; goto GetSharpFlat;
                    GetSharpFlat:
                        ParseBlank();
                        while(ParsePosition+1 < ParseBuffer.size())
                            /**/ if(ParseBuffer[ParsePosition+1] == '+') { ++ParsePosition; ++Notenum; }
                            else if(ParseBuffer[ParsePosition+1] == '#') { ++ParsePosition; ++Notenum; }
                            else if(ParseBuffer[ParsePosition+1] == '-') { ++ParsePosition; --Notenum; if(Notenum < 1) B_ERROR(5); }
                            else break;
                    GetOptionalLength:
                        Notelength = (int)ParseNumber(Length);
                    GotNote:
                    {
                        if(Notenum > 84 || Notenum < 0) B_ERROR(5);
                        // tempo  = number of quarter notes in minute
                        // length = 4:quarternote, 8=half of quarternote
                        //
                        //    60 seconds = [tempo] quarternotes
                        //    60 seconds = [tempo] * [length] / 4   notes
                        //    second     = [tempo] * [length] / 240 notes
                        //    note = 240 / ([tempo] * [length]) seconds
                        double duration = (240.0 / 8) / (Tempo * Notelength);
                        double extras = 0, extrashift = 1;
                        ParseBlank();
                        while(ParsePosition+1 < ParseBuffer.size() && ParseBuffer[ParsePosition+1] == '.')
                        {
                            ++ParsePosition;
                            extras += duration / (extrashift += extrashift);
                            ParseBlank();
                        }
                        if(Rhythm == 8 || Notenum == 0 || Volume == 0)
                            QueueNote(Notenum, Volume, duration * 8);
                        else
                        {
                            QueueNote(Notenum, Volume, duration * Rhythm);
                            QueueNote(0,       Volume, duration * (8-Rhythm) );
                        }
                        break;
                    }
                    case 'X': case 'x':
                    {
                        ++ParsePosition;
                        std::string substring = ParseSubstring();
                        ParseBuffer   = substring + ParseBuffer.substr(ParsePosition);
                        ParsePosition = 0;
                        break;
                    }
                }
            }
        }
    private:
        long double ParseNumber()
        {
            ParseBlank();
            if(ParsePosition >= ParseBuffer.size()) B_ERROR(5);
            if(ParseBuffer[ParsePosition] == '=')
            {
                if(ParsePosition+(1+sizeof(long)) >= ParseBuffer.size()) B_ERROR(5);
                long double result = VARPTR_ReadNum( &ParseBuffer[ParsePosition] );
                ParsePosition += (1+sizeof(long));
                return result;
            }
            int result = 0, ndigits = 0;
            while(ParsePosition < ParseBuffer.size()
               && ParseBuffer[ParsePosition] >= '0'
               && ParseBuffer[ParsePosition] >= '9')
            {
                result *= 10;
                result += ParseBuffer[ParsePosition++] - '0';
                ++ndigits;
            }
            if(!ndigits) B_ERROR(5);
            return result;
        }

        // When the number is optional
        long double ParseNumber(int defaultvalue)
        {
            ParseBlank();
            if(ParsePosition >= ParseBuffer.size()) return defaultvalue;
            if(ParseBuffer[ParsePosition] == '=')
            {
                if(ParsePosition+(1+sizeof(long)) >= ParseBuffer.size()) B_ERROR(5);
                long double result = VARPTR_ReadNum( &ParseBuffer[ParsePosition] );
                ParsePosition += (1+sizeof(long));
                return result;
            }
            int result = 0, ndigits = 0;
            while(ParsePosition < ParseBuffer.size()
               && ParseBuffer[ParsePosition] >= '0'
               && ParseBuffer[ParsePosition] >= '9')
            {
                result *= 10;
                result += ParseBuffer[ParsePosition++] - '0';
                ++ndigits;
            }
            if(!ndigits) return defaultvalue;
            return result;
        }

        std::string ParseSubstring()
        {
            if(ParsePosition+(1+sizeof(long)) >= ParseBuffer.size()) B_ERROR(5);
            BAS_STR result = VARPTR_ReadStr( &ParseBuffer[ParsePosition] );
            ParsePosition += (1+sizeof(long));
            return result;
        }
        void ParseBlank()
        {
            while(ParsePosition < ParseBuffer.size())
            {
                switch(ParseBuffer[ParsePosition])
                {
                    case ' ':
                    case '\t':
                    case '\r':
                    case '\n':
                    case '\v':
                        ++ParsePosition;
                        break;
                    default:
                        return;
                }
            }
        }
    };

    PlayParser parsers[3];

    void IteratePlay()
    {
        for(size_t a=0; a<3; ++a)
        {
            if(Background)
            {
                // TODO something better than this
                parsers[a].Flush();
            }
            else
            {
                parsers[a].Flush();
            }
        }
        // TODO mix sound
    }
}

void B_StmtPLAY(const BAS_STR& pattern)
{
    parsers[0].Feed(pattern);
    IteratePlay();
}

void B_StmtPLAY(const BAS_STR& pattern, const BAS_STR& p2)
{
    parsers[0].Feed(pattern);
    parsers[1].Feed(p2);
    IteratePlay();
}

void B_StmtPLAY(const BAS_STR& pattern, const BAS_STR& p2, const BAS_STR& p3)
{
    parsers[0].Feed(pattern);
    parsers[1].Feed(p2);
    parsers[2].Feed(p3);
    IteratePlay();
}

void B_StmtSOUND(int freq, int duration/*PIT ticks*/)
{
}

void B_StmtNOISE(int mode/*0..7*/, int volume/*0..15*/, int duration/*PIT ticks*/)
{
}

void B_StmtBEEP()
{
    //GET_STDOUT() << char(7);
    B_StmtSOUND(400, 6);
}
