#ifndef bqtSymTablesHH
#define bqtSymTablesHH

#include "autoptr"
#include <map>
#include <set>
#include <vector>
#include <string>
#include <utility>
#include <unordered_map>

typedef autoptr<struct TypeInfo> TypeP;
typedef autoptr<struct UserType> UtypeP;
typedef autoptr<struct Scope> ScopeP;
typedef autoptr<struct SubRoutine> SubFuncP;

extern const TypeP TypeInt, TypeSng, TypeDbl, TypeLng, TypeStr;
extern const TypeP TypeByte, TypeUbyte, TypeUint, TypeUlong;

enum class TypeType
{
    Single, Double, String, UserType,
    Byte,         Integer,         Long,
    UnsignedByte, UnsignedInteger, UnsignedLong
};
constexpr unsigned NumTypes = (4+3+3);

struct TypeInfo: public ptrable
{
    TypeType Type;

    enum ArrayType
    {
        NotArray,
        StaticArray,      // '$STATIC
        DynamicArray,     // '$DYNAMIC
        ParamArray        // One passed via parameter (number of dimensions unknown)
    };

    ArrayType arraytype;     //if array
    unsigned fixlength;      //if non-dynamic string.
    std::string utypename;   //if UserType
    UtypeP members;          //if UserType

    // If array:
    std::vector<std::pair<long, unsigned> > dimensions;
    // ^first  = beginning index
    // ^second = dimension (size)
    //   example: 1 TO 5   is (1,5)
    //            0 TO 4   is (0,5)

    TypeInfo(TypeType t = TypeType::Integer)
    : Type(t),arraytype(NotArray),fixlength(0),members(),dimensions()
    { }

    bool is_string() const { return Type == TypeType::String; }
    bool is_num() const    { return Type != TypeType::String
                                 && Type != TypeType::UserType; }
    bool is_array() const  { return arraytype != NotArray; }
    bool is_integer() const { return is_num() && !is_array()
                                  && Type != TypeType::Single
                                  && Type != TypeType::Double; }
    bool is_unsigned() const { return Type == TypeType::UnsignedByte
                                   || Type == TypeType::UnsignedLong
                                   || Type == TypeType::UnsignedInteger; }

    bool operator==(const TypeInfo& b) const;
    bool operator!=(const TypeInfo& b) const { return !operator==(b); }

    const std::string Dump() const;
};

#include "expr.hh"

struct UserType: public std::vector<std::pair<std::string, TypeP> >, public ptrable
{
    std::unordered_map<std::string, unsigned> Names;

    bool DeclareMember(const std::string& name, TypeP p);
    TypeP GetMember(const std::string& name) const;

    bool operator==(const UserType& b)const;
};

struct TypeFlagSet
{
    //union {
        struct{
    bool is_subroutine:1;
    bool is_scalar    :1;
    bool is_const     :1;
    bool is_array     :1;

    bool is_global_scalar : 1; // imported using SHARED
    bool is_global_array  : 1; // imported using SHARED

    // In case of subroutines, they can be DECLARED and later defined.
    bool has_body     :1;
        };
        //unsigned zero : 8;
    //};

    // Types and labels not included here; they are in different namespace.

    TypeFlagSet() :
        is_subroutine(0), is_scalar(0),
        is_const(0), is_array(0),
        is_global_scalar(0), is_global_array(0),
        has_body(0) {}
};

struct SubParamUsageReport
{
    struct ParamUsage
    {
        bool AssignedTo;
        std::set<std::pair<SubFuncP, unsigned/*param index*/> > PassedAsParam;

        ParamUsage(): AssignedTo(false),PassedAsParam() {}
    };
    /* for each param, their status */
    std::vector<ParamUsage> report;
};

/* SUB, FUNCTION, DEF FN */
struct SubRoutine: public ptrable
{
    struct ParamSpec
    {
        std::string name;
        TypeP       type;
        bool        is_array;
        enum { ByRef, ByVal, Seg } passconvention;
        bool        was_modified; // If the param is never modified,
                                  // It can be modified into ByVal.

        ParamSpec() : name(), type(), is_array(false),
          passconvention(ByRef), was_modified(false) { }

        void SetByval() { passconvention=ByVal; }
        void SetSeg() { passconvention=Seg; }
    };

    TypeP returntype; // 0=SUB, otherwise FUNCTION/DEF
    std::vector<ParamSpec> params;
    bool is_static;
    autoptr<class stmt> code;

    int callconvention;     // 0=default, 1=cdecl
    std::string linkername; // ALIAS

    ScopeP sub_scope;

    SubParamUsageReport params_usage;

public:
    SubRoutine(): returntype(), params(), is_static(false),
                  code(), callconvention(0), linkername(),
                  sub_scope(), params_usage() {}
};

struct LabelDef
{
    bool     defined;
    unsigned dataptr;
};

typedef autoptr<struct VarContextDef> ctxp;
struct VarContextDef: public ptrable
{
    TypeP type;
    bool is_static; // STATIC to the given function
    bool is_shared; // SHARED to the subfunctions (towards children)
    int param_index;
    //stmtp first_use;

public:
    VarContextDef(): type(),is_static(false),is_shared(false), param_index(-1)
    //               ,first_use(0)
    {}
};

typedef std::unordered_map<std::string, ctxp> ScopeVarListType;

struct Scope: public ptrable
{
    /*****
      Method of lookup:
        Check constants. Ignore suffix.
          Match? Fail if had wrong suffix. Otherwise, accept.

        Check UTYPE.
          Match? Fail if had a suffix. Otherwise, accept.
        Had suffix? Check by suffix.

        Otherwise, check if in Utypes.
          If the symbol was declared using DIM...AS,
          this gets the explicit type. DEFtype is the
          last resort.

        Otherwise, check by DEFtype.
    *******/

    std::unordered_map<std::string, UtypeP> types;    // Types have no suffix.
    std::unordered_map<std::string, exprp> constants; // Constants have no suffix.

    // SUBs, FUNCTIONs, DEFs
    std::unordered_map<std::string, SubFuncP> subroutines;

    ScopeVarListType Scopes[NumTypes][2]; // indexed by TypeType. Second index: 0=scalars, 1=arrays

    std::unordered_map<std::string, TypeFlagSet> usage;

    // Labels... Original name -> translated name
    std::unordered_map<std::string, std::string> labelnames;
    std::map<std::string, LabelDef> labels;
};

#endif
